import logging
import tempfile

from ..utils_xnat import upload_file
from ..utils_redcap import download_file


logger = logging.getLogger('garjus.automations.edat_redcap2xnat')


def process_project(
    xnat,
    project,
    events,
    tab_field,
    event2sess,
    scans,
    xnat_scan_type,
    xnat_scan_resource,
):
    results = []
    def_field = project.def_field
    fields = [def_field, tab_field]
    id2subj = {}

    # Get the subject id map
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}
    else:
        rec = project.export_records(fields=[def_field])
        id2subj = {x[def_field]: x[def_field] for x in rec if x[def_field]}

    # Get mri records
    rec = project.export_records(fields=fields, events=events)

    # Process each record
    for r in rec:
        record_id = r[def_field]
        event = r['redcap_event_name']

        try:
            subj = id2subj[record_id]
        except KeyError as err:
            logger.debug(f'record without subj num:{record_id}:{err}')
            continue

        logger.debug(f'{record_id}:{event}:{subj}')

        # Check for converted file
        if not r[tab_field]:
            logger.debug(f'{record_id}:{subj}:{event}:not yet converted')
            continue

        # Check if it needs uploading
        try:
            session = subj + event2sess[event]
            logger.debug(f'{session=}, {event2sess=}')
        except KeyError:
            logger.debug(f'{subj}:{event}:event not in map:{event2sess=}')
            continue

        # Filter scans for this session
        sess_scans = [x for x in scans if x['SESSION'] == session]
        sess_scans = [x for x in sess_scans if x['SCANTYPE'] == xnat_scan_type]

        # Filter scans to exclude unusable
        sess_scans = [x for x in sess_scans if x['QUALITY'] != 'unusable']

        if len(sess_scans) <= 0:
            logger.debug(f'scan not found:{subj}:{event}:{session}')
            continue

        if len(sess_scans) > 1:
            logger.debug(f'multiple scans, uhhh:{subj}:{event}:{session}')
            continue

        # Only one scan found so use it
        scan = sess_scans[0]
        scan_label = scan['SCANID']

        # Check for existing resource
        if xnat_scan_resource in scan['RESOURCES']:
            logger.debug(f'{session}:{scan_label}:{xnat_scan_resource}:exists')
            continue

        # Copy the file
        with tempfile.TemporaryDirectory() as tmpdir:
            # Download the tab file from redcap to tmp
            tab_file = f'{tmpdir}/{r[tab_field]}'
            result = download_file(
                project,
                record_id,
                tab_field,
                tab_file,
                event_id=event)

            if not result:
                logger.error(f'{subj}:{event}:download failed')
                continue

            resource = xnat.select_scan_resource(
                scan['PROJECT'],
                scan['SUBJECT'],
                scan['SESSION'],
                scan['SCANID'],
                xnat_scan_resource)

            # Upload file to xnat resource
            result = upload_file(tab_file, resource)

            if not result:
                logger.error(f'{subj}:{session}:{scan_label}:upload fail')
                continue

        logger.info(f'{subj}:{event}:{session}:{scan_label}:uploaded')
        results.append({
            'result': 'COMPLETE',
            'description': 'edat_redcap2xnat',
            'category': 'edat_redcap2xnat',
            'subject': subj,
            'session': session,
            'scan': scan['SCANID'],
            'event': event,
            'field': tab_field})

    return results



from garjus import Garjus


def get_stats(xnat, fullpath):
    filename = xnat.select(
        fullpath).resource('1stLEVEL').file('behavior.txt').get()
    data = {}
    rows = []

    try:
        with open(filename) as f:
            rows = f.readlines()

        for r in rows:
            (k, v) = r.strip().replace('=', ',').split(',')
            data[k] = v
    except ValueError:
        print(f'cannot load stats file:{filename}')
        return {}

    return data


g = Garjus()

df = g.assessors(projects=['REMBRANDT'])

df = df[df.PROCTYPE == 'fmri_msit_v3']

df = df.sort_values('ASSR').reset_index()

df = df[df['PROCSTATUS'] == 'COMPLETE']
df = df[df['QCSTATUS'] != 'Failed']

for i, row in df.iterrows():
    if i != 341:
        continue

    print(i, row['full_path'])
    try:
        stats = get_stats(g.xnat(), row['full_path'])
    except Exception as err:
        print(err)
        continue

    print('set stats', stats)
    g.set_stats(
        row['PROJECT'],
        row['SUBJECT'],
        row['SESSION'],
        row['ASSR'],
        stats)

print('DONE!')



"""ARC app."""
import logging
import tempfile

import pandas as pd

from ...utils_redcap import field2events, download_file


logger = logging.getLogger('garjus.automations.etl_arcapp')


def process_project(project):
    '''project is a pycap project for project that contains arcapp data'''
    results = []
    file_field = 'arcapp_csvfile'
    def_field = project.def_field
    fields = [def_field, file_field]
    id2subj = {}
    events = field2events(project, file_field)
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}
    else:
        rec = project.export_records(fields=[def_field])
        id2subj = {x[def_field]: x[def_field] for x in rec if x[def_field]}

    # Get records
    rec = project.export_records(fields=fields, events=events)

    # Process each record
    for r in rec:
        record_id = r[def_field]
        event_id = r['redcap_event_name']
        subj = id2subj.get(record_id)

        # Check for converted file
        if not r[file_field]:
            logging.debug(f'no arcapp file:{record_id}:{subj}:{event_id}')
            continue

        if r[file_field] == 'CONVERT_FAILED.txt':
            logging.debug(f'found CONVERT_FAILED')
            continue

        if r[file_field] == 'MISSING_DATA.txt':
            logging.debug(f'found MISSING_DATA')
            continue

        # Do the ETL
        result = _process(project, record_id, event_id)
        if result:
            results.append({
                'result': 'COMPLETE',
                'category': 'etl_arcapp',
                'description': 'etl_arcapp',
                'subject': subj,
                'event': event_id,
                'field': file_field})

    return results

def _process(project, record_id, event_id):
    file_field = 'arcapp_csvfile'

    logger.debug(f'etl_arcapp:{record_id}:{event_id}')

    # check for existing records in the repeating instruments for this event
    rec = project.export_records(
        records=[record_id],
        events=[event_id],
        forms=['arc_data'],
        fields=[project.def_field]
    )

    rec = [x for x in rec if x['redcap_repeat_instrument'] == 'arc_data']

    if len(rec) > 0:
        logger.debug(f'found existing records:{record_id}:{event_id}')
        return

    with tempfile.TemporaryDirectory() as tmpdir:
        csv_file = f'{tmpdir}/{record_id}-{event_id}-{file_field}.csv'

        logger.debug(f'downloading file:{record_id}:{event_id}')    

        try:
            download_file(
                project,
                record_id,
                file_field,
                csv_file,
                event_id=event_id)
        except Exception as err:
            logging.error(f'download failed:{record_id}:{event_id}:{err}')
            return False

        # Transform data
        try:
            data = _transform(csv_file)
            for d in data:
                d[project.def_field] = record_id
                d['redcap_event_name'] = event_id
        except Exception as err:
            logging.error(f'transform failed:{record_id}:{event_id}:{err}')
            import traceback
            traceback.print_exc()
            return False

    if not data:
        return False

    # Also complete the file form
    data.append({
        project.def_field: record_id,
        'redcap_event_name': event_id,
        'arc_app_complete': '2'
    })

    # Finally load back to redcap
    result = _load(project, data)

    return result


def _load(project, data):
    # Load the data back to redcap
    try:
        _response = project.import_records(data)
        assert 'count' in _response
        return True
    except (AssertionError, Exception) as err:
        logger.error(err)
        return False

def _read(filename):
    df = pd.DataFrame()

    try:
        # Load Data
        df = pd.read_excel(filename, dtype=str)
    except ValueError as err:
        logger.error(f'failed to read excel:{filename}:{err}')

    return df

def _transform(filename):
    data = []

    # Load the data
    logging.info(f'loading:{filename}')
    df = _read(filename)

    if df.empty:
        logging.debug(f'empty file')
        return []

    df = df.fillna('')    

    if df is None:
        logging.error('extract failed')
        return

    for i in range(1,29):
        r = None
        d = {
            'redcap_repeat_instance': str(i),
            'redcap_repeat_instrument': 'arc_data',
            'arc_order_index': str(i),
            'arc_data_complete': '2'}

        rows = df[df.id == str(i)]

        if len(rows) > 1:
            logger.error(f'multiple rows for id:{i}')
            return []
        elif len(rows) == 0:
            logger.debug(f'no rows for id:{i}')
            continue

        # Get the only record as dictionary
        r = rows.to_dict('records')[0]

        if r.get('completeTime', False):
            if 'T' in r['completeTime']:
                d['arc_completetime'] = r['completeTime'].split('T')[1].split('.')[0].rsplit(':', 1)[0]
            else:
                d['arc_completetime'] = r['completeTime'].split(' ')[1].split('.')[0].rsplit(':', 1)[0]

        if r.get('startTime', False):
            if 'T'  in r['startTime']:
                d['arc_response_date'] = r['startTime'].split('T')[0]
                d['arc_starttime'] = r['startTime'].split('T')[1].split('.')[0].rsplit(':', 1)[0]
            else:
                d['arc_response_date'] = r['startTime'].split(' ')[0]
                d['arc_starttime'] = r['startTime'].split(' ')[1].split('.')[0].rsplit(':', 1)[0]

        if r.get('prescribedTime', False):
            d['arc_notification_time'] = r['prescribedTime'].split('T')[1].split('.')[0].rsplit(':', 1)[0]

        if 'index' in r:
            d['arc_index_no'] = r['index']

        if 'dayIndex' in r:
            d['arc_day_index'] = r['dayIndex']

        if 'finishedSession' in r:
            d['arc_finishedsession'] = r['finishedSession'].lower().replace('true', '1').replace('false', '0')

        if 'interrupted' in r:
            d['arc_interrupted'] = r['interrupted'].lower().replace('true', '1').replace('false', '0')

        if 'symbolsRT' in r:
            d['arc_symbolsrt'] = r['symbolsRT']

        if 'symbolsAcc' in r:
            d['arc_symbols_accuracy'] = r['symbolsAcc']

        if 'pricesAcc' in r:
            d['arc_prices_accuracy'] = r['pricesAcc']

        if 'pricesRT' in r:
            d['arc_pricesrt'] = r['pricesRT']

        if 'gridEd' in r:
            d['arc_grided'] = r['gridEd']

        data.append(d)

    return data



"""Creates report PDF with zip."""
import logging
import io
import re
import os
import shutil
import itertools
from datetime import datetime, date, timedelta
import tempfile
import math

import numpy as np
import pydot
import pandas as pd
import plotly
import plotly.graph_objs as go
import plotly.subplots
import plotly.express as px
from fpdf import FPDF
from fpdf.enums import XPos, YPos
from PIL import Image


logger = logging.getLogger('garjus.progress.report')


# These are used to set colors of graphs
RGB_DKBLUE = 'rgb(59,89,152)'
RGB_BLUE = 'rgb(66,133,244)'
RGB_GREEN = 'rgb(15,157,88)'
RGB_YELL = 'rgb(244,160,0)'
RGB_RED = 'rgb(219,68,55)'
RGB_PURP = 'rgb(160,106,255)'
RGB_GREY = 'rgb(200,200,200)'
RGB_PINK = 'rgb(255,182,193)'
RGB_LIME = 'rgb(17, 180, 101)'

# Give each status a color to display
QASTATUS2COLOR = {
    'PASS': RGB_GREEN,
    'NQA': RGB_LIME,
    'NPUT': RGB_YELL,
    'FAIL': RGB_RED,
    'NONE': RGB_GREY,
    'JOBF': RGB_PINK,
    'JOBR': RGB_BLUE}

STATUS2RGB = dict(zip(
    ['WAITING', 'PENDING', 'RUNNING', 'COMPLETE', 'FAILED', 'UNKNOWN', 'JOBF'],
    [RGB_GREY, RGB_YELL, RGB_GREEN, RGB_BLUE, RGB_RED, RGB_PURP, RGB_PINK]))

# These are used to make progress reports
ASTATUS2COLOR = {
    'PASS': RGB_GREEN,
    'NPUT': RGB_YELL,
    'FAIL': RGB_RED,
    'NQA': RGB_LIME,
    'NONE': RGB_GREY,
    'COMPLETE': RGB_BLUE,
    'UNKNOWN': RGB_PURP}

SESSCOLS = ['SESSION', 'SUBJECT', 'PROJECT', 'DATE', 'SESSTYPE', 'SITE', 'MODALITY']

HIDECOLS = [
    'assessor_label',
    'PROJECT',
    'SESSION',
    'SUBJECT',
    'AGE',
    'SEX',
    'DEPRESS',
    'TYPE',
    'SITE',
    'SESSTYPE',
    'DATE',
    'ASSR',
    'PROCSTATUS',
    'QCSTATUS',
    'INPUTS',
    'MODALITY',
    'XSITYPE',
    'PROCTYPE',
    'full_path',
    'case',
]

ACOLS = [
    'ASSR',
    'PROJECT',
    'SUBJECT',
    'SESSION',
    'SESSTYPE',
    'SITE',
    'DATE',
    'PROCTYPE',
]


class MYPDF(FPDF):
    """Custom PDF."""

    def set_filename(self, filename):
        """Set the filename."""
        self.filename = filename

    def set_project(self, project, enable_monthly=False):
        """Set the project name."""
        self.project = project
        today = datetime.now().strftime("%Y-%m-%d")
        self.date = today

        if enable_monthly:
            self.title = f'{project} Monthly Report'
            self.subtitle = '{}'.format(datetime.now().strftime("%B %Y"))
        else:
            self.title = f'{project} Report'
            self.subtitle = '{}'.format(datetime.now().strftime("%B %d, %Y"))
            

    def footer(self):
        """Return the custom footer."""
        self.set_y(-0.35)
        self.set_x(0.5)

        # Write date, title, page number
        self.set_font('helvetica', size=10)
        self.set_text_color(100, 100, 100)
        self.set_draw_color(100, 100, 100)
        self.line(x1=0.2, y1=10.55, x2=8.3, y2=10.55)
        self.cell(w=1, text=self.date)
        self.cell(w=5, align='C', text=self.title)
        self.cell(w=2.5, align='C', text=str(self.page_no()))


def blank_letter():
    """Blank letter sized PDF."""
    p = MYPDF(orientation="P", unit='in', format='letter')
    p.set_top_margin(0.5)
    p.set_left_margin(0.5)
    p.set_right_margin(0.5)

    return p


def _draw_counts(pdf, sessions, rangetype=None, groupby='site'):
    # Counts of each session type with sums
    # sessions column names are: SESSION, PROJECT, DATE, SESSTYPE, SITE
    type_list = sessions.SESSTYPE.unique()
    site_list = sorted(sessions.SITE.unique())
    group_list = sessions.GROUP.unique()
    indent_width = max(2.5 - len(type_list) * 0.5, 0.3)

    # Get the data
    df = sessions.copy()

    if rangetype == 'lastmonth':
        pdf.set_fill_color(114, 172, 77)

        # Get the dates of lst month
        _end = date.today().replace(day=1) - timedelta(days=1)
        _start = date.today().replace(day=1) - timedelta(days=_end.day)

        # Get the name of last month
        lastmonth = _start.strftime("%B")

        # Filter the data to last month
        df = df[df.DATE >= _start.strftime('%Y-%m-%d')]
        df = df[df.DATE <= _end.strftime('%Y-%m-%d')]

        # Create the lastmonth header
        _txt = 'Session Counts ({})'.format(lastmonth)

    else:
        pdf.set_fill_color(94, 156, 211)
        _txt = 'Session Counts (all)'

    # Draw heading
    pdf.set_font('helvetica', size=14)
    pdf.cell(w=7.5, h=0.5, text=_txt, align='C', border=0, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Header Formatting
    pdf.cell(w=1.0)
    pdf.set_text_color(245, 245, 245)
    pdf.set_line_width(0.01)
    _kwargs = {'w': 1.0, 'h': 0.7, 'border': 1, 'align': 'C', 'fill': True}

    # Column header for each type
    pdf.cell(indent_width)
    for cur_type in type_list:
        _txt = cur_type
        if len(_txt) > 6:
            pdf.set_font('helvetica', size=12)

        pdf.cell(**_kwargs, text=_txt)

    # Got to next line
    pdf.ln()

    # Row formatting
    pdf.set_fill_color(255, 255, 255)
    pdf.set_text_color(0, 0, 0)
    _kwargs = {'w': 1.0, 'h': 0.5, 'border': 1, 'align': 'C', 'fill': False}
    _kwargs_s = {'w': 1.0, 'h': 0.5, 'border': 1, 'align': 'C', 'fill': False}
    _kwargs_t = {'w': 0.7, 'h': 0.5, 'border': 1, 'align': 'C', 'fill': False}

    pdf.set_font('helvetica', size=18)

    if groupby == 'site':

        # Row for each site
        for cur_site in site_list:
            pdf.cell(w=indent_width)

            dfs = df[df.SITE == cur_site]
            _txt = cur_site

            if len(_txt) > 9:
                pdf.set_font('helvetica', size=11)

            pdf.cell(**_kwargs_s, text=_txt)

            pdf.set_font('helvetica', size=18)

            # Count each type for this site
            for cur_type in type_list:
                cur_count = str(len(dfs[dfs.SESSTYPE == cur_type]))
                pdf.cell(**_kwargs, text=cur_count)

            if len(type_list) > 1:
                # Total for site
                cur_count = str(len(dfs))
                pdf.cell(**_kwargs_t, text=cur_count, new_x=XPos.LMARGIN, new_y=YPos.NEXT)
            else:
                pdf.cell(**{'w': 1.0, 'h': 0.5, 'border': 0}, text='', new_x=XPos.LMARGIN, new_y=YPos.NEXT)

        if len(site_list) > 1:
            # TOTALS row
            pdf.cell(w=indent_width)
            pdf.cell(w=1.0)
            for cur_type in type_list:
                pdf.set_font('helvetica', size=18)
                cur_count = str(len(df[df.SESSTYPE == cur_type]))
                pdf.cell(**_kwargs, text=cur_count)

            # Grandtotal
            pdf.cell(**_kwargs_t, text=str(len(df)))

    else:
        # Row for each group
        for cur_group in group_list:
            pdf.cell(w=indent_width)

            dfg = df[df.GROUP == cur_group]

            # Show the group
            _txt = cur_group

            if len(_txt) > 6:
                pdf.set_font('helvetica', size=12)

            pdf.cell(**_kwargs_s, text=_txt)

            # Count each type for this group
            pdf.set_font('helvetica', size=18)
            for cur_type in type_list:
                cur_count = str(len(dfg[dfg.SESSTYPE == cur_type]))
                pdf.cell(**_kwargs, text=cur_count)

            if len(type_list) > 1:
                # Total for group
                cur_count = str(len(dfg))
                pdf.cell(**_kwargs_t, text=cur_count, new_x=XPos.LMARGIN, new_y=YPos.NEXT)
            else:
                #pdf.cell(w=0, h=0, border=0, text='', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
                pdf.ln()

        if len(group_list) > 1:
            # TOTALS row
            pdf.cell(w=indent_width)
            pdf.cell(w=1.0)
            for cur_type in type_list:
                pdf.set_font('helvetica', size=18)
                cur_count = str(len(df[df.SESSTYPE == cur_type]))
                pdf.cell(**_kwargs, text=cur_count)

            # Grandtotal
            if len(type_list) > 1:
                pdf.cell(**_kwargs_t, text=str(len(df)))

    # End by going to next line
    pdf.ln()

    return pdf

def _exclude_maps(scantypes):
    return [x for x in scantypes if not 'FIELDMAP' in x.replace('_', '').upper() and not 'FSA' in x.replace('_', '').upper()]

def _draw_scan_counts(pdf, scans, groupby='site'):
    # Counts of each scan type by sess type by site/group
    scantypes = sorted(scans.SCANTYPE.unique())
    sesstypes = sorted(scans.SESSTYPE.unique())
    sitetypes = sorted(scans.SITE.unique())
    grouptypes = sorted(scans.GROUP.unique())

    scantypes = _exclude_maps(scantypes)

    if groupby == 'site':
        indent_width = max(2.0 - len(sitetypes) * 0.5, 0.3)
    else:
        indent_width = max(2.0 - len(grouptypes) * 0.5, 0.3)

    # Get the data
    df = scans.copy()

    pdf.set_fill_color(94, 156, 211)
    _txt = 'Scan Counts'

    # Draw heading
    pdf.set_font('helvetica', size=14)
    pdf.cell(w=7.5, h=0.5, text=_txt, align='C', border=0, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Header Formatting
    pdf.cell(w=1.0)
    pdf.set_text_color(245, 245, 245)
    pdf.set_line_width(0.01)

    if len(sitetypes) > 4:
        _kwargs = {'w': 0.5, 'h': 0.5, 'border': 1, 'align': 'C', 'fill': True}
    else:
        _kwargs = {'w': 1.0, 'h': 0.5, 'border': 1, 'align': 'C', 'fill': True}

    # Column header for each type 

    if groupby == 'site':
        if len(sitetypes) > 4:
            pdf.cell(w=indent_width + 1.2)
        else:
            pdf.cell(w=indent_width + 2.0)

        for cur_type in sitetypes:
            _txt = cur_type
            if len(_txt) > 6:
                pdf.set_font('helvetica', size=12)

            if len(sitetypes) > 4:
                _txt = _txt[:4]

            pdf.cell(**_kwargs, text=_txt)
    
    else:
        pdf.cell(w=indent_width + 2.0)

        for cur_type in grouptypes:
            _txt = cur_type
            if len(_txt) > 6:
                pdf.set_font('helvetica', size=12)

            pdf.cell(**_kwargs, text=cur_type)

    # Next line
    pdf.ln()

    # Row formatting
    pdf.set_fill_color(255, 255, 255)
    pdf.set_text_color(0, 0, 0)

    if len(sitetypes) > 4:
        _kwargs = {'w': 0.5, 'h': 0.4, 'border': 1, 'align': 'C', 'fill': False}
        _kwargs_s = {'w': 0.5, 'h': 0.4, 'border': 1, 'align': 'C', 'fill': False}
        _kwargs_x = {'w': 1.5, 'h': 0.4, 'border': 1, 'align': 'C', 'fill': False}
        _kwargs_t = {'w': 0.7, 'h': 0.4, 'border': 1, 'align': 'C', 'fill': False}
    else:
        _kwargs = {'w': 1.0, 'h': 0.4, 'border': 1, 'align': 'C', 'fill': False}
        _kwargs_s = {'w': 1.0, 'h': 0.4, 'border': 1, 'align': 'C', 'fill': False}
        _kwargs_x = {'w': 2.0, 'h': 0.4, 'border': 1, 'align': 'C', 'fill': False}
        _kwargs_t = {'w': 1.0, 'h': 0.4, 'border': 1, 'align': 'C', 'fill': False}

    pdf.set_font('helvetica', size=12)

    # First grouping by scan type
    #print(f'{indent_width=}')
    for cur_scan in scantypes:
        # Reset our cursor and indent
        pdf.cell(**{'w': 0, 'h': 0, 'border': 0}, text='', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
        pdf.cell(w=(indent_width))

        # Show the scan type
        pdf.set_font('helvetica', size=12, style='B')
        _txt = cur_scan
        pdf.cell(**_kwargs_x, text=_txt)

        # Row for each session type
        for cur_sess in sesstypes:
            # Show the session type
            pdf.set_font('helvetica', size=12)
            _txt = cur_sess
            #pdf.cell(**_kwargs_s, text=_txt)
            pdf.cell(**_kwargs_t, text=_txt)

            pdf.set_font('helvetica', size=16)
            if groupby == 'site':
                # Show count for each site
                for cur_site in sitetypes:
                    _txt = str(len(df[(df.SESSTYPE == cur_sess) & (df.SCANTYPE == cur_scan) & (df.SITE == cur_site)]))
                    pdf.cell(**_kwargs_s, text=_txt)
            else:
                # Show count for each group
                for cur_group in grouptypes:
                    _txt = str(len(df[(df.SESSTYPE == cur_sess) & (df.SCANTYPE == cur_scan) & (df.GROUP == cur_group)]))
                    pdf.cell(**_kwargs, text=_txt)

            # Fill then next line and indent
            pdf.cell(**{'w': 1.0, 'h': 0.4, 'border': 0}, text='', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
            if len(sitetypes) > 4:
                pdf.cell(w=indent_width + 1.5)
            else:
                pdf.cell(w=indent_width + 2.0)

    # End by going to next line
    pdf.ln()

    return pdf


def plot_timeline(df, startdate=None, enddate=None):
    """Plot timeline of data."""
    palette = itertools.cycle(px.colors.qualitative.Plotly)
    type_list = df.SESSTYPE.unique()
    mod_list = df.MODALITY.unique()
    fig = plotly.subplots.make_subplots(rows=1, cols=1)
    fig.update_layout(margin=dict(l=40, r=40, t=40, b=40))

    for mod, sesstype in itertools.product(mod_list, type_list):
        # Get subset for this session type
        dfs = df[(df.SESSTYPE == sesstype) & (df.MODALITY == mod)]
        if dfs.empty:
            continue

        # Advance color here, before filtering by time
        _color = next(palette)

        if startdate:
            dfs = dfs[dfs.DATE >= startdate.strftime('%Y-%m-%d')]

        if enddate:
            dfs = dfs[dfs.DATE <= enddate.strftime('%Y-%m-%d')]

        # Nothing to plot so go to next session type
        if dfs.empty:
            logger.debug('nothing to plot:{}:{}'.format(mod, sesstype))
            continue

        # markers symbols, see https://plotly.com/python/marker-style/
        if mod == 'MR':
            symb = 'circle-dot'
        elif mod == 'PET':
            symb = 'diamond-wide-dot'
        else:
            symb = 'diamond-tall-dot'

        # Convert hex to rgba with alpha of 0.5
        if _color.startswith('#'):
            _rgba = 'rgba({},{},{},{})'.format(
                int(_color[1:3], 16),
                int(_color[3:5], 16),
                int(_color[5:7], 16),
                0.7)
        else:
            _r, _g, _b = _color[4:-1].split(',')
            _a = 0.7
            _rgba = 'rgba({},{},{},{})'.format(_r, _g, _b, _a)

        # Plot this session type
        try:
            _row = 1
            _col = 1
            fig.append_trace(
                go.Box(
                    name='{} {} ({})'.format(sesstype, mod, len(dfs)),
                    x=dfs['DATE'],
                    y=dfs['SITE'],
                    boxpoints='all',
                    jitter=0.7,
                    text=dfs['SESSION'],
                    pointpos=0.5,
                    orientation='h',
                    marker={
                        'symbol': symb,
                        'color': _rgba,
                        'size': 12,
                        'line': dict(width=2, color=_color)
                    },
                    line={'color': 'rgba(0,0,0,0)'},
                    fillcolor='rgba(0,0,0,0)',
                    hoveron='points',
                ),
                _row,
                _col)
        except Exception as err:
            logger.error(err)
            return None

    # show lines so we can better distinguish categories
    fig.update_yaxes(showgrid=True)

    # Set the size
    fig.update_layout(width=900)

    # Export figure to image
    _png = fig.to_image(format="png")
    image = Image.open(io.BytesIO(_png))
    return image


def plot_activity(df, pivot_index):
    """Plot activity data."""
    status2rgb = ASTATUS2COLOR

    fig = plotly.subplots.make_subplots(rows=1, cols=1)
    fig.update_layout(margin=dict(l=40, r=40, t=40, b=40))

    # Draw bar for each status, these will be displayed in order
    dfp = pd.pivot_table(
        df, index=pivot_index, values='ID', columns=['STATUS'],
        aggfunc='count', fill_value=0)

    for status, color in status2rgb.items():
        ydata = sorted(dfp.index)
        if status not in dfp:
            xdata = [0] * len(dfp.index)
        else:
            xdata = dfp[status]

        fig.append_trace(go.Bar(
            x=ydata,
            y=xdata,
            name='{} ({})'.format(status, sum(xdata)),
            marker=dict(color=color),
            opacity=0.9), 1, 1)

    # Customize figure
    fig['layout'].update(barmode='stack', showlegend=True, width=900)

    # Export figure to image
    _png = fig.to_image(format="png")
    image = Image.open(io.BytesIO(_png))
    return image


def _add_count_pages(pdf, sessions, enable_monthly=False, groupby='site'):
    mr_sessions = sessions[sessions.MODALITY == 'MR'].copy()

    # Start the page with titles
    pdf.add_page()
    pdf.set_font('helvetica', size=22)
    pdf.cell(w=7.5, h=0.4, align='C', text=pdf.title, new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.cell(w=7.5, h=0.4, align='C', text=pdf.subtitle, border='B', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.ln(0.25)

    # Show all MRI session counts
    pdf.set_font('helvetica', size=18)
    pdf.cell(w=7.5, h=0.4, align='C', text=f'MRI by {groupby}')
    pdf.ln(0.25)
    _draw_counts(pdf, mr_sessions, groupby=groupby)
    pdf.ln(1)

    if enable_monthly:
        if len(mr_sessions.SITE.unique()) > 3:
            # Start a new page so it fits
            pdf.add_page()

        # Show MRI session counts in date range
        pdf.cell(w=7.5, h=0.4, align='C', text='MRI')
        pdf.ln(0.25)
        _draw_counts(pdf, mr_sessions, rangetype='lastmonth', groupby=groupby)
        pdf.ln(1)
        pdf.add_page()

    # Add other Modalities, counts for each session type
    logger.debug('adding other counts')
    _add_others(pdf, sessions, enable_monthly=enable_monthly, groupby=groupby)

    return pdf

def _add_scan_count_pages(pdf, scans, groupby='site'):
    mr_scans = scans[scans.MODALITY == 'MR'].copy()

    # Start the page
    pdf.add_page()

    # Show all MRI scan counts
    pdf.set_font('helvetica', size=18)
    pdf.cell(w=7.5, h=0.4, align='C', text=f'MRI by {groupby}')
    pdf.ln(0.25)
    _draw_scan_counts(pdf, mr_scans, groupby=groupby)
    pdf.ln(1)

    return pdf


def _add_graph_page(pdf, info):
    scantypes = info['scantypes']
    proctypes = info['proctypes']

    pdf.add_page()
    pdf.set_font('helvetica', size=18)
    pdf.cell(w=7.5, align='C', text='Processing Graph', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.set_font('helvetica', size=9)

    # MR Scan are orange
    pdf.set_fill_color(255, 166, 0)
    pdf.cell(h=0.3, text='MR Scan', fill=True, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # PET Scan are chocolate
    pdf.set_fill_color(210, 105, 30)
    pdf.cell(h=0.3, text='PET Scan', fill=True, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # EDAT are pink
    pdf.set_fill_color(238, 130, 238)
    pdf.cell(h=0.3, text='EDAT', fill=True, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Processing with stats are green
    pdf.set_fill_color(144, 238, 144)
    pdf.cell(h=0.3, text='Processing with stats', fill=True, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Processing without stats are blue
    pdf.set_fill_color(173, 216, 230)
    pdf.cell(h=0.3, text='Processing without stats', fill=True, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    pdf.ln(0.5)

    # Build the graph
    graph = pydot.Dot(graph_type='digraph')
    graph.set_node_defaults(
        color='lightblue',
        style='filled',
        shape='box',
        fontname='Courier',
        fontsize='12')

    for scan in scantypes:
        if scan == 'CTAC':
            if 'FEOBVQA_v2' in proctypes:
                graph.add_node(pydot.Node('FEOBV', color='chocolate'))

            if 'AMYVIDQA_v2' in proctypes:
                graph.add_node(pydot.Node('AMYVID', color='chocolate'))
        else:
            graph.add_node(pydot.Node(scan, color='orange'))

    print(proctypes)

    if 'centiloids_pib_v1' in proctypes:
        graph.add_node(pydot.Node('PiB', color='chocolate'))
        graph.add_node(pydot.Node('AMYLOIDQA_v4', color='lightgreen'))
        graph.add_node(pydot.Node('centiloids_pib_v1', color='lightgreen'))
        graph.add_edge(pydot.Edge('PiB', 'AMYLOIDQA_v4'))
        graph.add_edge(pydot.Edge('FS7_v1', 'AMYLOIDQA_v4'))
        graph.add_edge(pydot.Edge('PiB', 'centiloids_pib_v1'))
        graph.add_edge(pydot.Edge('T1', 'centiloids_pib_v1'))

    if 'fMRI_NBACK' in scantypes:
        graph.add_node(pydot.Node('NBACK', color='violet'))

    if 'NMQA_v1' in proctypes:
        graph.add_node(pydot.Node('NMQA_v1', color='lightblue'))
        graph.add_edge(pydot.Edge('NM', 'NMQA_v1'))
        graph.add_edge(pydot.Edge('T1', 'NMQA_v1'))

    if 'fMRI_EEfRT1' in scantypes:
        graph.add_node(pydot.Node('EEfRT', color='violet'))
        graph.add_node(pydot.Node('fmriqa_v4', color='lightgreen'))
        graph.add_edge(pydot.Edge('fMRI_EEfRT1', 'fmriqa_v4'))
        graph.add_edge(pydot.Edge('fMRI_EEfRT2', 'fmriqa_v4'))
        graph.add_edge(pydot.Edge('fMRI_EEfRT3', 'fmriqa_v4'))

    if 'fMRI_MIDT1' in scantypes:
        graph.add_node(pydot.Node('MIDT', color='violet'))
        graph.add_edge(pydot.Edge('fMRI_MIDT1', 'fmriqa_v4'))
        graph.add_edge(pydot.Edge('fMRI_MIDT2', 'fmriqa_v4'))

    # Default proctypes
    graph.add_node(pydot.Node('FS7_v1', color='lightgreen'))
    graph.add_node(pydot.Node('FS7HPCAMG_v1', color='lightgreen'))
    graph.add_node(pydot.Node('LST_v1', color='lightgreen'))
    graph.add_node(pydot.Node('SAMSEG_v1', color='lightgreen'))
    graph.add_node(pydot.Node('DnSeg_v1', color='lightgreen'))
    graph.add_node(pydot.Node('FS7sclimbic_v0', color='lightgreen'))
    graph.add_node(pydot.Node('FS7hypothal_v1', color='lightgreen'))

    graph.add_edge(pydot.Edge('T1', 'DnSeg_v1'))
    graph.add_edge(pydot.Edge('T1', 'LST_v1'))
    graph.add_edge(pydot.Edge('T1', 'FS7_v1'))
    graph.add_edge(pydot.Edge('FS7_v1', 'SAMSEG_v1'))
    graph.add_edge(pydot.Edge('FS7_v1', 'FS7HPCAMG_v1'))
    graph.add_edge(pydot.Edge('FLAIR', 'LST_v1'))
    graph.add_edge(pydot.Edge('FLAIR', 'SAMSEG_v1'))
    graph.add_edge(pydot.Edge('FS7_v1', 'FS7sclimbic_v0'))
    graph.add_edge(pydot.Edge('FS7_v1', 'FS7hypothal_v1'))

    if 'fmri_msit_v4' in proctypes:
        graph.add_node(pydot.Node('EDAT', color='violet'))
        graph.add_edge(pydot.Edge('EDAT', 'fmri_msit_v4'))
        graph.add_edge(pydot.Edge('T1', 'fmri_msit_v4'))
        graph.add_node(pydot.Node('fmri_msit_v4', color='lightgreen'))
        graph.add_edge(pydot.Edge('fMRI_MSIT', 'fmri_msit_v4'))

    if 'fmri_bct_v2' in proctypes:
        graph.add_node(pydot.Node('struct_preproc_v1', color='lightgreen'))
        graph.add_edge(pydot.Edge('T1', 'struct_preproc_v1'))
        graph.add_edge(pydot.Edge('FLAIR', 'struct_preproc_v1'))
        graph.add_edge(pydot.Edge('struct_preproc_v1', 'fmri_rest_v4'))
        graph.add_edge(pydot.Edge('fMRI_REST1', 'fmri_rest_v4'))
        graph.add_edge(pydot.Edge('fMRI_REST2', 'fmri_rest_v4'))
        graph.add_edge(pydot.Edge('fmri_rest_v4', 'fmri_roi_v2'))
        graph.add_edge(pydot.Edge('fmri_roi_v2', 'fmri_bct_v2'))
        graph.add_node(pydot.Node('fmri_bct_v2', color='lightgreen'))

    if 'struct_preproc_noflair_v1' in proctypes:
        graph.add_edge(pydot.Edge(
            'struct_preproc_noflair_v1', 'fmri_rest_v4', style='dashed'))
        graph.add_edge(pydot.Edge(
            'T1', 'struct_preproc_noflair_v1', style='dashed'))

    if 'BFC_v2' in proctypes:
        graph.add_edge(pydot.Edge('T1', 'BFC_v2'))
        graph.add_node(pydot.Node('BFC_v2', color='lightgreen'))

    if 'FEOBVQA_v2' in proctypes:
        graph.add_edge(pydot.Edge('FEOBV', 'FEOBVQA_v2'))
        graph.add_edge(pydot.Edge('FS7_v1', 'FEOBVQA_v2'))
        graph.add_node(pydot.Node('FEOBVQA_v2', color='lightgreen'))

    if 'AMYVIDQA_v2' in proctypes:
        graph.add_edge(pydot.Edge('AMYVID', 'AMYVIDQA_v2'))
        graph.add_edge(pydot.Edge('FS7_v1', 'AMYVIDQA_v2'))
        graph.add_node(pydot.Node('AMYVIDQA_v2', color='lightgreen'))

    if 'BrainAgeGap_v2' in proctypes:
        graph.add_node(pydot.Node('Multi_Atlas_v3', color='lightgreen'))
        graph.add_node(pydot.Node('BrainAgeGap_v2', color='lightgreen'))
        graph.add_edge(pydot.Edge('T1', 'Multi_Atlas_v3'))
        graph.add_edge(pydot.Edge('Multi_Atlas_v3', 'BrainAgeGap_v2'))

    if 'dtiQA_synb0_v7' in proctypes:
        graph.add_node(pydot.Node('dtiQA_synb0_v7', color='lightgreen'))
        graph.add_node(pydot.Node('francois_schaefer200_v1', color='lightblue'))
        graph.add_edge(pydot.Edge('T1', 'dtiQA_synb0_v7'))
        graph.add_edge(pydot.Edge('DTI_2min_b1000', 'dtiQA_synb0_v7'))
        graph.add_edge(pydot.Edge('DTI_2min_b2000', 'dtiQA_synb0_v7'))
        graph.add_edge(pydot.Edge('dtiQA_synb0_v7', 'francois_schaefer200_v1'))
        graph.add_edge(pydot.Edge('FS7_v1', 'francois_schaefer200_v1'))

    # Make the graph, draw to pdf
    image = Image.open(io.BytesIO(graph.create_png()))
    pdf.image(image, x=0.5, y=3, w=7.5)

    return pdf


def _add_others(pdf, sessions, enable_monthly=False, groupby='site'):
    # Get non-MRI sessions
    other_sessions = sessions[sessions.MODALITY != 'MR'].copy()

    if len(other_sessions) == 0:
        logger.debug('no other modalities sessions, skipping page')
        return

    # Show all session counts
    pdf.set_font('helvetica', size=18)
    pdf.cell(w=7.5, h=0.4, align='C', text=f'Other Modalities by {groupby}')
    pdf.ln(0.25)
    _draw_counts(pdf, other_sessions, groupby=groupby)
    pdf.ln(1)

    if enable_monthly:
        # Show session counts in date range
        pdf.cell(w=7.5, h=0.4, align='C', text='Other Modalities')
        pdf.ln(0.25)
        _draw_counts(pdf, other_sessions, rangetype='lastmonth', groupby=groupby)
        pdf.ln(1)

    return pdf


def _add_wml_page(pdf, info):

    # Transform the data to row per session with columsn for LST, SAMSEG wml
    stats = info['stats']

    if stats.empty:
        logger.info('no stats found')
        return

    lst_data = stats[stats.PROCTYPE == 'LST_v1']
    sam_data = stats[stats.PROCTYPE == 'SAMSEG_v1']

    if lst_data.empty:
        logger.info('no LST data')
        return

    if sam_data.empty:
        logger.info('no SAMSEG data')
        return

    df = pd.merge(lst_data, sam_data, left_on='SESSION', right_on='SESSION')

    pdf.add_page()
    pdf.cell(text='LST vs SAMSEG', align='C', new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    fig = plotly.subplots.make_subplots(rows=1, cols=1)

    fig.append_trace(
        go.Scatter(
            x=df['wml_volume_x'].astype(float),
            y=df['samseg_lesions_y'].astype(float) / 1000.0,
            mode='markers',
        ), 1, 1)

    _max = max(
        df['wml_volume_x'].astype(float).max(),
        df['samseg_lesions_y'].astype(float).max() / 1000.0,
    )
    _max = max(_max, 50)

    fig.add_trace(
        go.Scatter(
            x=[0, _max],
            y=[0, _max],
            mode='lines',
            name='',
            line_color='grey'), 1, 1)

    fig.update_yaxes(autorange=True)
    fig.update_layout(showlegend=False)
    fig.update_layout(
        xaxis_title="LST wml (mL)",
        yaxis_title="SAMSEG lesions (mL)")

    # Draw to figure as image on PDF
    _image = Image.open(io.BytesIO(fig.to_image(format="png")))
    pdf.image(_image, x=0.75, w=7)

    return pdf


def _add_stats_fmriqa(pdf, stats, info):
    # get the stats data by scan type using inputs field to map to scan

    scans = info['scans']

    assessors = info['assessors'].copy()
    assessors = assessors[assessors.PROCTYPE == 'fmriqa_v4']

    # Extract scan fmri value into a column
    assessors['scan_fmri'] = assessors.apply(
        lambda row: row['INPUTS'].get('scan_fmri'), axis=1)

    df = pd.merge(
        assessors[['ASSR', 'scan_fmri']],
        scans[['full_path', 'SCANTYPE']],
        how='inner',
        left_on='scan_fmri',
        right_on='full_path')

    df = pd.merge(
        stats,
        df[['ASSR', 'SCANTYPE']],
        how='inner',
        left_on='ASSR',
        right_on='ASSR')

    for t in sorted(df.SCANTYPE.unique()):
        if t in ['fMRI_rest', 'fMRI_REST_FSA']:
            continue

        pdf.set_font('helvetica', size=10)
        _add_stats(pdf, df[df.SCANTYPE == t], plot_title=f'Scan Type: {t}')


def _add_stats(pdf, stats, plot_title=None):

    # this returns a PIL Image object
    image = plot_stats(stats, plot_title)
    tot_width, tot_height = image.size

    # Split horizontal image into chunks of width to fit on
    # letter-sized page with crop((left, top, right, bottom))
    chunk_h = 500
    chunk_w = 998
    rows_per_page = 3  # 3 rows per page
    page_count = math.ceil(tot_width / (rows_per_page * chunk_w))

    for p in range(page_count):
        for c in range(rows_per_page):
            # Calculate the starting x for this chunk
            chunk_x = (c * chunk_w) + (p * chunk_w * rows_per_page)
            if chunk_x > tot_width - 10:
                # out of bounds
                continue

            # Get the image from the cropped section
            _img = image.crop((chunk_x, 0, chunk_x + chunk_w, chunk_h))

            # Draw the image on the PDF
            pdf.image(_img, x=0.75, h=3.1)

    return pdf


def _add_qa_page(pdf, scandata, assrdata, sesstype):
    scan_image = plot_qa(scandata)
    assr_image = plot_qa(assrdata)

    if not scan_image and not assr_image:
        # skip this page b/c there's nothing to plot
        logger.debug('skipping page, nothing to plot:{}'.format(sesstype))
        return pdf

    pdf.add_page()
    pdf.set_font('helvetica', size=18)
    pdf.ln(0.5)
    pdf.cell(w=5, align='C', text='Scans by Type ({} Only)'.format(sesstype))

    if scan_image:
        pdf.image(scan_image, x=0.5, y=1.3, w=7.5)
        pdf.ln(4.7)

    if assr_image:
        pdf.cell(w=5, align='C', text=f'Assessors by Type ({sesstype} Only)')
        pdf.image(assr_image, x=0.5, y=6, w=7.5)

    return pdf


def _add_timeline_page(pdf, info, enable_monthly=False):
    # Get the data for all
    df = info['sessions'].copy()

    pdf.add_page()

    pdf.set_font('helvetica', size=18)

    # Draw all timeline
    _txt = 'Sessions Timeline'
    if enable_monthly:
        _txt += ' (all)'

    pdf.cell(w=7.5, align='C', text=_txt)
    image = plot_timeline(df)
    pdf.image(image, x=0.5, y=0.75, w=7.5)
    pdf.ln(5)

    if enable_monthly:
        # Get the dates of last month
        enddate = date.today().replace(day=1) - timedelta(days=1)
        startdate = date.today().replace(day=1) - timedelta(days=enddate.day)

        # Get the name of last month
        lastmonth = startdate.strftime("%B")
        _txt = 'Sessions Timeline ({})'.format(lastmonth)
        image = plot_timeline(df, startdate=startdate, enddate=enddate)
        pdf.cell(w=7.5, align='C', text=_txt)
        pdf.image(image, x=0.5, y=5.75, w=7.5)
        pdf.ln()
        pdf.add_page()

    return pdf


def _add_nda_page(pdf, info):
    pdf.add_page()

    pdf.cell(text='NDA', new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    pdf.ln(1)

    return pdf

def _add_analyses_page(pdf, info):
    pdf.add_page()

    pdf.cell(text='ANALYSES', new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    analyses = info['analyses']

    #print(analyses)

    for i, a in analyses.iterrows():
        #PROJECT
        #ID
        #NAME
        #STATUS
        #EDIT
        #NOTES
        #SUBJECTS
        #PROCESSOR
        #INVESTIGATOR
        #OUTPUT

        _txt = f'{a.PROJECT}-A{a.ID} {a.NAME}'
        pdf.cell(text=_txt, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

        _txt = f'STATUS:{a.STATUS} OUTPUT:{a.OUTPUT}'
        pdf.cell(text=_txt, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

        _txt = f'NOTES:{a.NOTES}'
        pdf.cell(text=_txt, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

        pdf.ln(0.3)


    return pdf

def _add_settings_page(pdf, info):
    pdf.add_page()

    pdf.set_font('helvetica', size=18)
    pdf.cell(w=7.5, align='C', text='Project Settings', new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Add some space
    pdf.ln(0.5)

    # Main Settings
    pdf.set_font('helvetica', size=14, style='B')
    pdf.cell(text='REDCap Projects:', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.set_font('helvetica', size=10)
    pdf.cell(h=0.2, text=f'Primary PID:{info["primary_redcap"]}', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.cell(h=0.2, text=f'Secondary PID:{info["secondary_redcap"]}', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.cell(h=0.2, text=f'Stats PID:{info["stats_redcap"]}', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.ln(0.3)

    if info.get('xnat_scanmap', False):
        # Scan Map
        pdf.set_font('helvetica', size=14, style='B')
        pdf.cell(text='XNAT Scan Map', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
        pdf.set_font('helvetica', size=10)
        _txt = _transform_scanmap(info['xnat_scanmap'])
        pdf.multi_cell(w=5.0, h=0.2, text=_txt, new_x=XPos.LMARGIN, new_y=YPos.NEXT)
        pdf.ln(0.3)

    if info.get('nda_scanmap', False):
        # NDA Scan Map
        pdf.set_font('helvetica', size=14, style='B')
        pdf.cell(text='NDA Scan Map', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
        pdf.set_font('helvetica', size=11)
        _txt = info['nda_scanmap']
        pdf.multi_cell(w=5.0, h=0.2, text=_txt, new_x=XPos.LMARGIN, new_y=YPos.NEXT)
        pdf.ln(0.3)

    if info.get('nda_expmap', False):
        # NDA Experiment Map
        pdf.set_font('helvetica', size=14, style='B')
        pdf.cell(text='NDA Experiment Map', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
        pdf.set_font('helvetica', size=11)
        _txt = info['nda_expmap']
        pdf.multi_cell(w=5.0, h=0.2, text=_txt, new_x=XPos.LMARGIN, new_y=YPos.NEXT)
        pdf.ln(0.3)

    if info.get('scan_protocols', False):
        # Scanning Protocols
        pdf.set_font('helvetica', size=14, style='B')
        pdf.cell(text='Scanning Protocols', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
        pdf.set_font('helvetica', size=11)
        _txt = '\n'.join([x['scanning_eventname'] for x in info['scan_protocols']])
        pdf.multi_cell(w=5.0, h=0.2, text=_txt, new_x=XPos.LMARGIN, new_y=YPos.NEXT)
        pdf.ln(0.3)

    if info.get('edat_protocols', False):
        # EDAT Protocols
        pdf.set_font('helvetica', size=14, style='B')
        pdf.cell(text='EDAT Protocols', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
        pdf.set_font('helvetica', size=11)
        _txt = '\n'.join([x['edat_name'] for x in info['edat_protocols']])
        pdf.multi_cell(w=5.0, h=0.2, text=_txt, new_x=XPos.LMARGIN, new_y=YPos.NEXT)
        pdf.ln(0.2)

    return pdf


def _add_proclib_page(pdf, info):
    pdf.add_page()

    # Get the proclib for enabled proctypes
    proclib = info['proclib']
    proclib = {k: v for k, v in proclib.items() if k in info['proctypes']}

    # Display each proctype
    for k, v in proclib.items():
        # Show the proctype
        pdf.set_font('helvetica', size=16)
        pdf.cell(text=k, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

        # Build the description
        _text = v.get('short_descrip', '') + '\n'
        _text += 'Inputs: ' + v.get('inputs_descrip', '') + '\n'
        _text += v.get('procurl', '') + '\n'

        # show stats
        for s, t in info['statlib'].get(k, {}).items():
            _text += f'{s}: {t}\n'

        # Show the description
        pdf.set_font('helvetica', size=12)
        pdf.multi_cell(0, 0.3, _text, border='LBTR', align="L", new_x=XPos.RIGHT, new_y=YPos.NEXT)

        # Add some space between proc types
        pdf.ln(0.2)

    return pdf


def _add_phantoms(pdf, info, enable_monthly=False):
    # Get the data for all
    df = info['phantoms'].copy()

    # Draw all timeline
    _txt = 'Phantoms (all)'
    pdf.set_font('helvetica', size=18)
    pdf.cell(w=7.5, align='C', text=_txt, new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    image = plot_timeline(df)
    pdf.image(image, x=0.5, w=7.5)
    pdf.ln(5)

    if enable_monthly:
        # Get the dates of last month
        enddate = date.today().replace(day=1) - timedelta(days=1)
        startdate = date.today().replace(day=1) - timedelta(days=enddate.day)

        # Get the name of last month
        lastmonth = startdate.strftime("%B")

        _txt = 'Phantoms ({})'.format(lastmonth)
        image = plot_timeline(df, startdate=startdate, enddate=enddate)
        pdf.cell(w=7.5, align='C', text=_txt)
        pdf.image(image, x=0.5, y=5.75, w=7.5)
        pdf.ln()

    return pdf


def _add_activity_page(pdf, info):
    pdf.add_page()
    pdf.set_font('helvetica', size=16)

    # top third is jobs section
    df = info['recentjobs'].copy()
    image = plot_activity(df, 'CATEGORY')
    pdf.image(image, x=1.6, y=0.2, h=3.3)
    pdf.ln(0.5)
    pdf.multi_cell(1.5, 0.3, text='Jobs\n')

    # middle third is activity section
    df = info['activity'].copy()
    image = plot_activity(df, 'CATEGORY')
    pdf.image(image, x=1.6, y=3.5, h=3.3)
    pdf.ln(3)
    pdf.multi_cell(1.5, 0.3, text='Autos')

    # bottom third is issues
    df = info['issues'].copy()
    image = plot_activity(df, 'CATEGORY')
    pdf.image(image, x=1.6, y=7.0, h=3.3)
    pdf.ln(3)
    pdf.multi_cell(1.5, 0.3, text='Issues\n')

    return pdf


def plot_qa(dfp):
    """Plot QA bars."""
    for col in dfp.columns:
        if col in ('SESSION', 'PROJECT', 'DATE', 'MODALITY'):
            # don't mess with these columns
            continue

        # Change each value from the multiple values in concatenated
        # characters to a single overall status
        dfp[col] = dfp[col].apply(get_metastatus)

    # Initialize a figure
    fig = plotly.subplots.make_subplots(rows=1, cols=1)
    fig.update_layout(margin=dict(l=40, r=40, t=40, b=40))

    # Check for empty data
    if len(dfp) == 0:
        logger.debug('dfp empty data')
        return None

    # use pandas melt function to unpivot our pivot table
    df = pd.melt(
        dfp,
        id_vars=(
            'SESSION',
            'SUBJECT',
            'PROJECT',
            'DATE',
            'SITE',
            'SESSTYPE',
            'MODALITY'),
        value_name='STATUS')

    # Check for empty data
    if len(df) == 0:
        logger.debug('df empty data')
        return None

    # We use fill_value to replace nan with 0
    dfpp = df.pivot_table(
        index='TYPE',
        columns='STATUS',
        values='SESSION',
        aggfunc='count',
        fill_value=0)

    scan_type = []
    assr_type = []
    for cur_type in dfpp.index:
        # Use a regex to test if name ends with _v and a number, then assr
        if re.search('_v\d+$', cur_type):
            assr_type.append(cur_type)
        else:
            scan_type.append(cur_type)

    newindex = scan_type + assr_type
    dfpp = dfpp.reindex(index=newindex)

    # Draw bar for each status, these will be displayed in order
    # ydata should be the types, xdata should be count of status
    # for each type
    for cur_status, cur_color in QASTATUS2COLOR.items():
        ydata = dfpp.index
        if cur_status not in dfpp:
            xdata = [0] * len(dfpp.index)
        else:
            xdata = dfpp[cur_status]

        cur_name = '{} ({})'.format(cur_status, sum(xdata))

        _width = (len(xdata) * 0.1) + 0.1
        _width = min(_width, 0.8)
        fig.append_trace(
            go.Bar(
                x=ydata,
                y=xdata,
                name=cur_name,
                marker=dict(color=cur_color),
                opacity=0.9,
                width=_width,
            ),
            1, 1)

    # Customize figure
    fig['layout'].update(barmode='stack', showlegend=True, width=900)

    # Export figure to image
    _png = fig.to_image(format="png")
    image = Image.open(io.BytesIO(_png))

    return image


def _plottable(var):
    try:
        _ = var.astype(str).replace('', np.nan).dropna().str.strip('%').astype(float)
        return True
    except Exception:
        return False


def plot_stats(df, plot_title=None):
    """Plot stats, one boxlplot per var."""
    box_width = 250
    min_box_count = 4

    logger.debug('plot_stats:{}'.format(len(df)))

    # Check for empty data
    if len(df) == 0:
        logger.debug('empty data, using empty figure')
        fig = go.Figure()
        _png = fig.to_image(format="png")
        image = Image.open(io.BytesIO(_png))
        return image

    # Filter var list to only include those that have data
    var_list = [x for x in df.columns if not pd.isnull(df[x]).all()]

    # Filter var list to only stats variables
    var_list = [x for x in var_list if x not in HIDECOLS]

    # Filter var list to only stats can be plotted as float
    var_list = [x for x in var_list if _plottable(df[x])]

    # skip vars
    var_list = [x for x in var_list if not x.endswith('pathlength')]

    # Determine how many boxplots we're making, depends on how many vars, use
    # minimum so graph doesn't get too small
    box_count = len(var_list)
    site_count = len(df['SITE'].unique())

    if box_count < 3:
        box_width = 500
        min_box_count = 2
    elif box_count < 6:
        box_width = 333
        min_box_count = 3

    if site_count > 3:
        box_width = 500
        min_box_count = 2

    if box_count < min_box_count:
        box_count = min_box_count

    graph_width = box_width * box_count

    # Horizontal spacing cannot be greater than (1 / (cols - 1))
    hspacing = 1 / (box_count * 4)

    logger.debug(f'{box_count}, {min_box_count}, {graph_width}, {hspacing}')

    # Make the figure with 1 row and a column for each var we are plotting
    var_titles = [x[:22] for x in var_list]
    fig = plotly.subplots.make_subplots(
        rows=1,
        cols=box_count,
        horizontal_spacing=hspacing,
        subplot_titles=var_titles)

    # Add box plot for each variable
    for i, var in enumerate(var_list):

        # Create boxplot for this var and add to figure
        logger.debug('plotting var:{}'.format(var))

        _row = 1
        _col = i + 1

        fig.append_trace(
            go.Box(
                y=df[var].astype(str).replace('', np.nan).dropna().str.strip('%').astype(float),
                x=df['SITE'],
                boxpoints='all',
                text=df['ASSR'],
                boxmean=True,
            ),
            _row,
            _col)

        # Plot horizontal line at median
        _median = df[var].astype(str).replace('', np.nan).dropna().str.strip('%').astype(float).median()
        fig.add_trace(
            go.Scatter(
                x=df['SITE'],
                y=[_median] * len(df),
                mode='lines',
                name='',
                fillcolor='red',
                line_color='grey'),
            _row, _col)

        fig.update_yaxes(autorange=True)

        if var.startswith('con_') or var.startswith('inc_'):
            logger.debug('setting beta range:{}'.format(var))
            _yaxis = 'yaxis{}'.format(i + 1)
            fig['layout'][_yaxis].update(range=[-1, 1], autorange=False)
        else:
            pass

    # Move the subtitles to bottom instead of top of each subplot
    if len(df['SITE'].unique()) < 4:
        for i in range(len(fig.layout.annotations)):
            fig.layout.annotations[i].update(y=-.15)

    if plot_title:
        fig.update_layout(
            title={'text': plot_title, 'x': 0.5, 'xanchor': 'center'})

    # Customize figure to hide legend and fit the graph
    fig.update_layout(
        showlegend=False,
        autosize=False,
        width=graph_width,
        margin=dict(l=20, r=40, t=40, b=80, pad=0))

    _png = fig.to_image(format="png")
    image = Image.open(io.BytesIO(_png))
    return image


def _add_stats_pages(pdf, info):
    proclib = info['proclib']
    stats = info['stats']
    stattypes = info['stattypes']

    for proctype in stattypes:
        # Limit the data to this proctype
        stat_data = stats[stats.PROCTYPE == proctype]

        if stat_data.empty:
            logger.debug(f'no stats for proctype:{proctype}')
            continue

        logger.debug(f'add stats page:{proctype}')

        # Get descriptions for this processing type
        proc_info = proclib.get(proctype, {})

        # use proclib to filter stats variable names
        _subset = proc_info.get('stats_subset', None)
        if _subset:
            try:
                stat_data = stat_data[_subset + ['SITE', 'ASSR']]
            except Exception as err:
                logger.error(f'subset failed:{_subset}, skipping page:{proctype}')
                return

        # Now make the page
        pdf.add_page()
        pdf.set_font('helvetica', size=16)
        pdf.cell(text=proctype, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

        if proctype == 'fmriqa_v4':
            # stats by scan type using inputs field to map to scan
            _add_stats_fmriqa(pdf, stat_data, info)
        else:
            _add_stats(pdf, stat_data)

        # Build the description
        _text = proc_info.get('short_descrip', '') + '\n'
        _text += 'Inputs: ' + proc_info.get('inputs_descrip', '') + '\n'

        # Append stats descriptions
        for s, t in info['statlib'].get(proctype, {}).items():
            _text += f'{s}: {t}\n'

        # Show the descriptions
        pdf.set_font('helvetica', size=12)
        pdf.multi_cell(0, 0.25, _text, border='LBTR', align="L", new_x=XPos.RIGHT, new_y=YPos.NEXT)

        _url = proc_info.get('procurl', '')
        if _url:
            pdf.set_font('helvetica', size=10)
            pdf.cell(text=_url, link=_url)


def make_pdf(info, filename):
    enable_monthly = info['enable_monthly']
    multi_group = info['multi_group']

    """Make PDF from info, save to filename."""
    logger.debug('making PDF')

    # Initialize a new PDF letter size and shaped
    pdf = blank_letter()
    pdf.set_filename(filename)
    pdf.set_project(info['project'], enable_monthly=enable_monthly)

    # Add first page showing MRIs
    logger.debug('adding first page')
    _add_count_pages(pdf, info['sessions'], enable_monthly=enable_monthly, groupby='site')

    # Show group pages only when multiple groups found
    if multi_group:
        _add_count_pages(pdf, info['sessions'], enable_monthly=enable_monthly, groupby='group')

    # Add per scan counts
    logger.debug('adding per scan count pages')
    _add_scan_count_pages(pdf, info['scans'], groupby='site')
    if multi_group:
        _add_scan_count_pages(pdf, info['scans'], groupby='group')

    # Timeline
    logger.debug('adding timeline page')
    _add_timeline_page(pdf, info, enable_monthly=enable_monthly)

    # Phantom pages
    if len(info['phantoms']) > 0:
        logger.debug('adding phantom page')
        _add_phantoms(pdf, info, enable_monthly=enable_monthly)
    else:
        logger.debug('no phantom page')

    # Add stats pages
    if info['stats'].empty:
        logger.debug('without stats')
    else:
        logger.debug('adding stats pages')
        _add_stats_pages(pdf, info)

    # Session type pages - counts per scans, counts per assessor
    logger.debug('adding MR qa pages')

    mr_sessions = info['sessions'].copy()
    mr_sessions = mr_sessions[mr_sessions.MODALITY == 'MR']

    for curtype in mr_sessions.SESSTYPE.unique():
        logger.debug('add_qa_page:{}'.format(curtype))

        # Get the scan and assr data
        scandf = info['scanqa'].copy()
        assrdf = info['assrqa'].copy()

        # Limit to the current session type
        scandf = scandf[scandf.SESSTYPE == curtype]
        assrdf = assrdf[assrdf.SESSTYPE == curtype]

        # Drop columns that are all empty
        scandf = scandf.dropna(axis=1, how='all')
        assrdf = assrdf.dropna(axis=1, how='all')

        # Add the page for this session type
        _add_qa_page(pdf, scandf, assrdf, curtype)

    # LST vs SAMSEG
    _add_wml_page(pdf, info)

    # QA/Jobs/Issues counts
    if info['enable_monthly']:
        _add_activity_page(pdf, info)

    # Directed Graph of processing
    # TODO: only run if graphviz/dot are installed
    # TODO: build the graph dynamically using same logic as dashboard autofilter
    # to find used scan types, then list unused to the side with counts
    # and then do the for proc types based on enabled in processing or not
    _add_graph_page(pdf, info)

    # Settings
    _add_settings_page(pdf, info)

    # NDA
    if False:
        _add_nda_page(pdf, info)

    # Analyses
    _add_analyses_page(pdf, info)


    # Save to file
    logger.debug('saving PDF to file:{}'.format(pdf.filename))
    try:
        pdf.output(pdf.filename)
    except Exception as err:
        logger.error('error while saving PDF:{}:{}'.format(pdf.filename, err))

    return True


def _scanqa(scans, scantypes=None):
    dfp = _scan_pivot(scans).reset_index()

    if not scantypes:
        scantypes = scans.SCANTYPE.unique()

    # Filter columns to include
    include_list = SESSCOLS.copy()
    if scantypes:
        include_list += scantypes
    include_list = [x for x in include_list if x in dfp.columns]
    include_list = list(set(include_list))
    dfp = dfp[include_list]

    # Drop columns that are all empty
    dfp = dfp.dropna(axis=1, how='all')

    return dfp


def _assrqa(assessors, proctypes=None):
    # Load that data
    dfp = _assr_pivot(assessors).reset_index()

    # Filter columns to include
    include_list = SESSCOLS + proctypes
    include_list = [x for x in include_list if x in dfp.columns]
    include_list = list(set(include_list))
    dfp = dfp[include_list]

    return dfp


def _scan_pivot(df):
    _index = ('SESSION', 'SUBJECT', 'PROJECT', 'DATE', 'SESSTYPE', 'SITE', 'MODALITY')
    df['TYPE'] = df['SCANTYPE']
    df['STATUS'] = df['QUALITY']
    dfp = df.pivot_table(
        index=_index,
        columns='TYPE',
        values='STATUS',
        aggfunc=lambda x: ''.join(x))

    return dfp


def _assr_pivot(df):
    _index = ('SESSION', 'SUBJECT', 'PROJECT', 'DATE', 'SESSTYPE', 'SITE', 'MODALITY')
    df['TYPE'] = df['PROCTYPE']
    df['STATUS'] = df['PROCSTATUS']
    dfp = df.pivot_table(
        index=_index,
        columns='TYPE',
        values='STATUS',
        aggfunc=lambda x: ''.join(x))

    return dfp


def get_metastatus(status):
    if not status or pd.isnull(status):  # np.isnan(status):
        # empty so it's none
        metastatus = 'NONE'
    elif 'P' in status:
        # at least one passed, so PASSED
        metastatus = 'PASS'
    elif 'Q' in status:
        # any are still needs qa, then 'NEEDS_QA'
        metastatus = 'NQA'
    elif 'N' in status:
        # if any jobs are still running, then NEEDS INPUTS?
        metastatus = 'NPUT'
    elif 'F' in status:
        # at this point if one failed, then they all failed, so 'FAILED'
        metastatus = 'FAIL'
    elif 'X' in status:
        metastatus = 'JOBF'
    elif 'R' in status:
        metastatus = 'JOBR'
    elif 'usable' in status:
        metastatus = 'PASS'
    elif 'questionable' in status:
        metastatus = 'NQA'
    elif 'unusable' in status:
        metastatus = 'FAIL'
    else:
        # whatever else is UNKNOWN, grey
        metastatus = 'NONE'

    return metastatus


def _filter_scantypes(scantypes):

    # Try to filter out junk
    scantypes = [x for x in scantypes if not x.startswith('[')]
    scantypes = [x for x in scantypes if 'survey' not in x.lower()]
    scantypes = [x for x in scantypes if x.lower() != 'cor']
    scantypes = [x for x in scantypes if x.lower() != 'unknown']
    scantypes = [x for x in scantypes if x.lower() != 'fmri_rest_fsa']
    scantypes = [x for x in scantypes if not x.lower().startswith('screen')]
    scantypes = [x for x in scantypes if not x.startswith('Low Dose CT')]
    scantypes = [x for x in scantypes if not x.startswith('VWIP')]
    scantypes = [x for x in scantypes if not x.startswith('3DFRP')]
    scantypes = [x for x in scantypes if not x.startswith('TOPUP')]
    scantypes = [x for x in scantypes if not x.startswith('localizer')]
    scantypes = [x for x in scantypes if not x.startswith('Calibration')]
    scantypes = [x for x in scantypes if not x.endswith('PhysioLog')]
    scantypes = [x for x in scantypes if not x.endswith('SBRef')]
    scantypes = [x for x in scantypes if not x.endswith('FSA')]
    scantypes = [x for x in scantypes if not x.startswith('Cor_')]
    scantypes = [x for x in scantypes if not x.startswith('Ax_')]
    scantypes = [x for x in scantypes if not x.startswith('AXIIAL')]
    scantypes = [x for x in scantypes if not x.startswith('DTI_1_')]
    scantypes = [x for x in scantypes if 'MDDW' not in x]
    scantypes = [x for x in scantypes if not x.startswith('3-Plane')]
    scantypes = [x for x in scantypes if not x.startswith('DTI_96d')]
    scantypes = [x for x in scantypes if not x.startswith('Head-Low')]
    scantypes = [x for x in scantypes if not x.startswith('MultiP')]
    scantypes = [x for x in scantypes if not x.startswith('MPRAGE A')]
    scantypes = [x for x in scantypes if not x.startswith('CTAC2mm')]
    scantypes = [x for x in scantypes if not x.startswith('ORIG')]
    scantypes = [x for x in scantypes if not x.startswith('Phoenix')]
    scantypes = [x for x in scantypes if not x.startswith('SpinEcho')]
    scantypes = [x for x in scantypes if not x.startswith('rsfMRI')]
    scantypes = [x for x in scantypes if not x.startswith('Sagittal_3D_F')]
    scantypes = [x for x in scantypes if not x.startswith('fMRI_rest')]
    scantypes = [x for x in scantypes if not x.startswith('DTI_2min_b1000a')]
    scantypes = [x for x in scantypes if not x.startswith('DTI_2min_b1000a')]
    scantypes = [x for x in scantypes if not x.startswith('DTI_b0')]

    return scantypes


def make_project_report(
    garjus,
    project,
    pdfname,
    zipname=None,
    monthly=False
):
    """"Make the project report PDF and zip files"""
    proclib = garjus.processing_library()
    statlib = garjus.stats_library()
    activity = garjus.activity(project)
    issues = garjus.issues(project)
    analyses = garjus.analyses([project], download=False)

    # Load types for this project
    proctypes = garjus.proctypes(project)
    scantypes = garjus.scantypes(project)
    stattypes = garjus.stattypes(project)

    # Loads scans/assessors with type filters applied
    scans = garjus.scans(projects=[project], scantypes=scantypes)
    scans.SESSTYPE = scans.SESSTYPE.replace('', 'UNKNOWN')
    scans.SITE = scans.SITE.replace('', 'UNKNOWN')
    scans.MODALITY = scans.MODALITY.replace('', 'UNKNOWN')
    scans.DATE = scans.DATE.fillna(datetime.now())

    scantypes = list(scans.SCANTYPE.unique())
    scantypes = _filter_scantypes(scantypes)

    # Truncate names for display
    scantypes = list(set([x[:15].strip() for x in scantypes if x]))
    scantypes = sorted(scantypes)

    assessors = garjus.assessors(projects=[project], proctypes=proctypes)
    assessors.SESSTYPE = assessors.SESSTYPE.replace('', 'UNKNOWN')
    assessors.SITE = assessors.SITE.replace('', 'UNKNOWN')
    assessors.MODALITY = assessors.MODALITY.replace('', 'UNKNOWN')
    assessors.DATE = assessors.DATE.fillna(datetime.now())

    phantoms = garjus.phantoms(project)
    phantoms = phantoms[SESSCOLS].drop_duplicates().sort_values('SESSION')

    # Extract sessions from scans/assessors
    sessions = pd.concat([scans[SESSCOLS], assessors[SESSCOLS]])
    sessions = sessions.drop_duplicates().sort_values('SESSION')

    # Merge in group from subjects
    subjects = garjus.subjects(project).reset_index()
    sessions = pd.merge(
        sessions,
        subjects[['ID', 'PROJECT', 'GROUP']],
        left_on=('SUBJECT', 'PROJECT'),
        right_on=('ID', 'PROJECT'),
        how='left',
    )

    if project == 'REMBRANDT':
        sessions['GROUP'] = sessions['GROUP'].fillna('Depress')
    else:
        sessions['GROUP'] = sessions['GROUP'].fillna('UNKNOWN')

    scans = pd.merge(
        scans,
        subjects[['ID', 'PROJECT', 'GROUP']],
        left_on=('SUBJECT', 'PROJECT'),
        right_on=('ID', 'PROJECT'),
        how='left'
    )

    if project == 'REMBRANDT':
        scans['GROUP'] = scans['GROUP'].fillna('Depress')
    else:
        scans['GROUP'] = scans['GROUP'].fillna('UNKNOWN')

    # Load stats with extra assessor columns
    stats = garjus.stats(project, assessors)

    # Make the info dictionary for PDF
    info = {}
    info['scans'] = scans
    info['assessors'] = assessors
    info['proclib'] = proclib
    info['statlib'] = statlib
    info['project'] = project
    info['stattypes'] = stattypes
    info['scantypes'] = scantypes
    info['proctypes'] = proctypes
    info['sessions'] = sessions
    info['activity'] = activity
    info['issues'] = issues
    info['analyses'] = analyses
    info['recentjobs'] = _recent_jobs(assessors)
    info['recentqa'] = _recent_qa(assessors)
    info['stats'] = stats
    info['scanqa'] = _scanqa(scans, scantypes)
    info['assrqa'] = _assrqa(assessors, proctypes)
    info['phantoms'] = phantoms
    info['enable_monthly'] = monthly
    info['xnat_scanmap'] = garjus.project_setting(project, 'scanmap')
    info['nda_expmap'] = garjus.project_setting(project, 'xst2nei')
    info['nda_scanmap'] = garjus.project_setting(project, 'xst2nst')
    info['scan_protocols'] = garjus.scanning_protocols(project)
    info['edat_protocols'] = garjus.edat_protocols(project)
    info['settings_redcap'] = garjus.project_setting(project, 'redcap')
    info['primary_redcap'] = garjus.project_setting(project, 'primary')
    info['secondary_redcap'] = garjus.project_setting(project, 'secondary')
    info['stats_redcap'] = garjus.project_setting(project, 'stats')

    if len(sessions['GROUP'].unique()) > 1:
        info['multi_group'] = True
    else:
        info['multi_group'] = False

    # Save the PDF report to file
    make_pdf(info, pdfname)

    # Save the stats to zip file
    if zipname:
        # TODO: include a QA.csv and a subjects.csv with demographics
        subjects = garjus.subjects(project)
        data2zip(subjects, stats, zipname)


def data2zip(subjects, stats, filename):
    """Convert stats dict to zip of csv files, one csv per proctype."""
    with tempfile.TemporaryDirectory() as tmpdir:
        # Prep output dir
        data_dir = os.path.join(tmpdir, 'data')
        zip_file = os.path.join(tmpdir, 'data.zip')
        os.mkdir(data_dir)

        # Save subjects csv
        csv_file = os.path.join(data_dir, f'subjects.csv')
        logger.info(f'saving subjects csv:{csv_file}')
        subjects.to_csv(csv_file)

        # Save a csv for each proc type
        for proctype in stats.PROCTYPE.unique():
            # Get the data for this processing type
            dft = stats[stats.PROCTYPE == proctype]

            dft = dft.dropna(axis=1, how='all')

            dft = dft.sort_values('ASSR')

            # Save file for this type
            csv_file = os.path.join(data_dir, f'{proctype}.csv')
            logger.info(f'saving csv:{proctype}:{csv_file}')
            dft.to_csv(csv_file, index=False)

        # Create zip file of dir of csv files
        shutil.make_archive(data_dir, 'zip', data_dir)

        # Save it outside of temp dir
        logger.info(f'saving zip:{filename}')
        shutil.copy(zip_file, filename)


def _last_month():
    from dateutil.relativedelta import relativedelta
    return (datetime.today() - relativedelta(months=1)).strftime('%Y-%m-%d')


def _recent_jobs(assessors, startdate=None):
    """Get recent jobs, assessors on XNAT with job date since startdate."""
    if startdate is None:
        startdate = _last_month()

    df = assessors.copy()

    # Filter by jobstartdate date, include anything with job running
    df = df[(df['JOBDATE'] >= startdate) | (df['PROCSTATUS'] == 'JOB_RUNNING')]

    # Relabel as jobs
    df['LABEL'] = df['ASSR']
    df['CATEGORY'] = df['PROCTYPE']
    df['STATUS'] = df['PROCSTATUS'].map({
        'COMPLETE': 'COMPLETE',
        'JOB_FAILED': 'FAIL',
        'JOB_RUNNING': 'NPUT'}).fillna('UNKNOWN')
    df['CATEGORY'] = df['PROCTYPE']
    df['DESCRIPTION'] = 'JOB' + ':' + df['LABEL']
    df['DATETIME'] = df['JOBDATE']
    df['ID'] = df.index

    return df


def _recent_qa(assessors, startdate=None):
    if startdate is None:
        startdate = _last_month()

    df = assessors.copy()

    # Filter by qc date
    df = df[df['QCDATE'] >= startdate]

    # Relabel as qa
    df['LABEL'] = df['ASSR']
    df['CATEGORY'] = df['PROCTYPE']
    df['STATUS'] = df['QCSTATUS'].map({
        'Failed': 'FAIL',
        'Passed': 'PASS'}).fillna('UNKNOWN')
    df['CATEGORY'] = df['PROCTYPE']
    df['DESCRIPTION'] = 'QA' + ':' + df['LABEL']
    df['DATETIME'] = df['QCDATE']
    df['ID'] = df.index

    return df


def _transform_scanmap(scanmap):
    """Parse scan map stored as string into map."""
    # Parse multiline string of delimited key value pairs into dictionary
    scanmap = dict(x.strip().split(':', 1) for x in scanmap.split('\n'))

    # Remove extra whitespace from keys and values
    scanmap = {k.strip(): v.strip() for k, v in scanmap.items()}
    scanmap = '\n'.join(f'{k} -> {v}' for k, v in scanmap.items())

    return scanmap



import logging
import pandas as pd


logger = logging.getLogger('garjus.automations.etl_nihtoolbox_drtaylor')


def process(reg_file, score_file):
    # Extract data from downloaded files
    reg_data = _extract_regdata(reg_file)
    score_data = _extract_cogscores(score_file)

    # Transform data to match desired field names
    data = _transform(reg_data, score_data)

    return data


def _transform(regdata, scoredata):
    # Initialize test data
    picseqtest = None
    listsorttest = None
    patterntest = None
    picvocabtest = None
    oralrecogtest = None
    cogcrystalcomp = None
    data = {}

    # Start with the registration data
    data.update({
        'toolbox_pin': regdata['PIN'],
        'toolbox_deviceid': regdata['DeviceID'],
        'toolbox_age': regdata['Age'],
        'toolbox_education': regdata['Education'],
        'toolbox_gender': regdata['Gender'],
        'toolbox_handedness': regdata['Handedness'],
        'toolbox_race': regdata['Race'],
        'toolbox_ethnicity': regdata['Ethnicity'],
        'toolbox_assessment': regdata['Assessment Name'],
    })

    # Find the Pic Seq data that has mutliple versions
    for i in list(scoredata.keys()):
        if i.startswith('NIH Toolbox Picture Sequence Memory Test'):
            picseqtest = scoredata[i]

    # Load the other tests
    listsorttest = scoredata['NIH Toolbox List Sorting Working Memory Test Age 7+ v2.1']
    patterntest = scoredata['NIH Toolbox Pattern Comparison Processing Speed Test Age 7+ v2.1']
    picvocabtest = scoredata['NIH Toolbox Picture Vocabulary Test Age 3+ v2.1']
    oralrecogtest = scoredata['NIH Toolbox Oral Reading Recognition Test Age 3+ v2.1']

    # Get the individual scores
    data.update({
        'toolbox_listsorttest_raw': listsorttest['RawScore'],
        'toolbox_patterntest_raw': patterntest['RawScore'],
        'toolbox_picseqtest_raw': picseqtest['RawScore'],
        'toolbox_oralrecogtest_theta': oralrecogtest['Theta'],
        'toolbox_picseqtest_theta': picseqtest['Theta'],
        'toolbox_picvocabtest_theta': picvocabtest['Theta'],
        'toolbox_listsorttest_uncstd': listsorttest['Uncorrected Standard Score'],
        'toolbox_oralrecogtest_uncstd': oralrecogtest['Uncorrected Standard Score'],
        'toolbox_patterntest_uncstd': patterntest['Uncorrected Standard Score'],
        'toolbox_picseqtest_uncstd': picseqtest['Uncorrected Standard Score'],
        'toolbox_picvocabtest_uncstd': picvocabtest['Uncorrected Standard Score'],
        'toolbox_listsorttest_agestd': listsorttest['Age-Corrected Standard Score'],
        'toolbox_oralrecogtest_agestd': oralrecogtest['Age-Corrected Standard Score'],
        'toolbox_patterntest_agestd': patterntest['Age-Corrected Standard Score'],
        'toolbox_picseqtest_agestd': picseqtest['Age-Corrected Standard Score'],
        'toolbox_picvocabtest_agestd': picvocabtest['Age-Corrected Standard Score'],
        'toolbox_listsorttest_tscore': listsorttest['Fully-Corrected T-score'],
        'toolbox_oralrecogtest_tscore': oralrecogtest['Fully-Corrected T-score'],
        'toolbox_patterntest_tscore': patterntest['Fully-Corrected T-score'],
        'toolbox_picseqtest_tscore': picseqtest['Fully-Corrected T-score'],
        'toolbox_picvocabtest_tscore': picvocabtest['Fully-Corrected T-score'],
    })

    cogcrystalcomp = scoredata.get('Cognition Crystallized Composite v1.1', None)
    #audlearntest = scoredata.get('NIH Toolbox Auditory Verbal Learning Test (Rey) Age 8+ v2.0', None)

    #if audlearntest:
    #    data.update({
    #        'toolbox_audlearntest_raw': audlearntest['RawScore'],
    #    })

    if cogcrystalcomp:
        data.update({
            'toolbox_cogcrystalcomp_uncstd': cogcrystalcomp['Uncorrected Standard Score'],
            'toolbox_cogcrystalcomp_agestd': cogcrystalcomp['Age-Corrected Standard Score'],
            'toolbox_cogcrystalcomp_tscore': cogcrystalcomp['Fully-Corrected T-score'],
        })

    return data


def _extract_regdata(filename):
    data = {}

    try:
        df = pd.read_csv(filename)
    except Exception:
        df = pd.read_excel(filename)

    # Get data from last row
    data = df.iloc[-1].to_dict()

    return data


def _extract_cogscores(filename):
    data = {}

    # Load csv
    try:
        df = pd.read_csv(filename)
    except Exception:
        df = pd.read_excel(filename)

    # Drop instrument duplicates, keeping the last only
    df = df.drop_duplicates(subset='Inst', keep='last')

    # convert to dict of dicts indexed by Instrument
    df = df.dropna(subset=['Inst'])
    df = df.set_index('Inst')
    data = df.to_dict('index')

    return data



import fnmatch
import re
import json
import logging


logger = logging.getLogger('garjus.automations.xnat_relabel_session')


SESS_URI = '/REST/experiments?xsiType=xnat:imagesessiondata&columns=\
ID,\
label,\
project,\
xsiType,\
subject_label,\
session_label,\
session_type,\
xnat:imagesessiondata/acquisition_site,\
xnat:imagesessiondata/date,\
modality'

# RELABELS
# If the key in [0] matches the value in [1],
# then change the value for key in [2] to the value in [3]
# if[0]==[1]: [2] = [3]


def process_project(xnat, project, relabels, replace):
    """Apply relabels to project sessions."""
    results = relabel_sessions(
        xnat, project, relabels, replace=replace, overwrite=True)

    return results


def relabel(xnat, session, relabels, overwrite=False, replace=[]):
    cur_sess = session['label']
    cur_proj = session['project']
    cur_subj = session['subject_label']
    allowed_k1 = ['session_type', 'session_label']
    allowed_k2 = ['session_type', 'site']
    mset = {}

    logger.debug(f'relabels={relabels}')
    logger.debug(f'replace={replace}')

    # test each relabel
    for k1, v1, k2, v2 in relabels:
        if k1 not in allowed_k1:
            logger.warning(f'not allowed:{k1}')
            continue

        if k2 not in allowed_k2:
            logger.warning(f'not allowed:{k2}')
            continue

        # Try to match
        if not re.match(fnmatch.translate(v1), session[k1]):
            # no match
            continue

        if (overwrite is False) or (session[k2] and (session[k2] not in replace)):
            # There's already a value there
            continue
        elif session[k2] == v2:
            # have we already set this to the same thing
            continue

        # Add the relabel to mset
        if k2 == 'site':
            # We have to use the full path for site, no shortcut
            mset['xnat:imagesessiondata/acquisition_site'] = v2
        else:
            mset[k2] = v2

    # Check for empty set
    if not mset:
        return None

    # Connect to the session on xnat and apply new values
    logger.info(f'{cur_proj}:{cur_sess}:setting:{mset}')
    sess_obj = xnat.select_session(cur_proj, cur_subj, cur_sess)
    sess_obj.attrs.mset(mset)

    result = {
        'description': 'xnat_relabel_session',
        'result': 'COMPLETE',
        'category': 'xnat_relabel_session',
        'subject': cur_subj,
        'session': cur_sess}

    return result


def relabel_sessions(xnat, project, relabels, overwrite=False, replace=[]):
    results = []

    # get a list of sessions from the project
    sess_uri = '{}&project={}'.format(SESS_URI, project)
    json_data = json.loads(xnat._exec(sess_uri, 'GET'), strict=False)
    sessions = json_data['ResultSet']['Result']

    # iterate and relabel as needed
    for sess in sessions:
        sess['site'] = sess['xnat:imagesessiondata/acquisition_site']
        result = relabel(xnat, sess, relabels, overwrite, replace)
        if result:
            results.append(result)

    return results



import glob
import logging

import redcap

from ...utils_redcap import upload_file


logger = logging.getLogger('garjus.automations.fitbit.file2redcap')


# TODO: get some other date from redcap and compare it with data in file

# Set the field

# Get the events from redcap for that field

# Enter a file matcher for each event?

# Iterate each redcap instance of the field and try to find a file or

# Iterate files and try to find a matching field? or both?


class File2Redcap():
    # map of redcap event to session number in the filename as saved
    event2sess = {
        'baselinemonth_0_arm_2': '1',
        'baselinemonth_0_arm_3': '1',
        'month_8_arm_2': '2',
        'month_8_arm_3': '2',
        'month_16_arm_2': '3',
        'month_16_arm_3': '3',
        'month_24_arm_2': '4',
        'month_24_arm_3': '4',
    }

    # name of field in redcap where file is to be uploaded
    file_field = 'fitbit_summary_worn'

    # initialize the uploader with a pycap project and a directory
    def __init__(self, rc, rootdir=''):
        self.rc = rc
        self.rootdir = rootdir
        self.load_secondary_id()

    def load_secondary_id(self):
        # Load secondary ID
        dfield = self.rc.def_field
        sfield = self.rc.export_project_info()['secondary_unique_field']
        rec = self.rc.export_records(fields=[dfield, sfield])
        self.id2subj = {x[dfield]: x[sfield] for x in rec if x[sfield]}

    def upload_data_file(self, record, event, filename):
        '''upload a file to specified record and event.'''
        upload_file(
            self.rc,
            record,
            self.file_field,
            filename,
            event
        )

    def get_session_files(self, subj, sess_num):
        '''Returns list of file paths for specified subject and session'''
        sess_glob = f'{self.rootdir}/{subj}/Fitbit/Session*{sess_num}*/TimeFBWorn.csv'
        file_list = sorted(glob.glob(sess_glob))
        return file_list

    def process_record(self, r):
        '''Process the given record dictionary.'''
        record_id = r[self.rc.def_field]
        event = r['redcap_event_name']
        sess_num = self.event2sess[event]

        try:
            subj = self.id2subj[record_id]
        except KeyError as err:
            logging.debug(f'record without subject number:{err}')
            return None

        # Check for existing file
        if r[self.file_field]:
            logging.debug(f'{subj}:{event}:already uploaded')
            return None

        # Find files for this subject/session in local folder
        file_list = self.get_session_files(subj, sess_num)
        file_count = len(file_list)
        if file_count <= 0:
            logging.debug(f'{subj}:{event}:{sess_num}:no files matched')
            return None
        elif file_count > 1:
            logging.error(f'{subj}:{event}:{sess_num}:too many files matched')
            return None

        # Upload the first and only file found
        data_file = file_list[0]
        try:
            logger.debug(f'uploading:{data_file}')
            self.upload_data_file(record_id, event, data_file)
        except(ValueError, redcap.RedcapError) as err:
            logging.error(f'error uploading:{data_file}:{err}')
            return None

        logging.info(f'{subj}:{event}:{self.file_field}:{data_file}:uploaded')

        return {
            'result': 'COMPLETE',
            'type': 'file2redcap',
            'subject': self.id2subj[record_id],
            'event': r['redcap_event_name'],
            'field': self.file_field}

    def load_records(self):
        # we want to load id and file from all events with files
        def_field = self.rc.def_field
        fields = [def_field, self.file_field]
        events = self.event2sess.keys()

        # Get records for those fields/events
        logging.info('export records from REDCap')
        records = self.rc.export_records(fields=fields, events=events)

        records = [x for x in records if self.id2subj.get(x[def_field], False)]

        records = sorted(records, key=lambda x: self.id2subj.get(x[def_field]))

        return records

    def run(self):
        results = []

        # Process each record
        for r in self.load_records():
            result = self.process_record(r)
            if result:
                results.append(result)

        return results



import glob
import logging
import json
import os
from datetime import datetime

from ...utils_redcap import upload_file


logger = logging.getLogger('garjus.automations.etl_arcdata.file2redcap')


# first load file to determine participant_id, session, day, session_date
# drop leading zero from participant_id

# Try to match with existing record. Upload to existing record.

# match to fields in redcap:
# subj_num = participant_id
# arc_response_date =  session_date
# arc_day_index = day
# arc_order_index = id


def _load_testfile(testfile):
    data = []

    with open(testfile) as f:
        data = json.load(f)

    return data


def _subject_files(subject, files):
    subj_files = []

    if subject.startswith('3REM'):
        subject_code = '03' + subject[4:]
    else:
        subject_code = subject

    for testfile in files:
        d = _load_testfile(testfile)
        if d['participant_id'][1:] == subject_code:
            subj_files.append(testfile)

    return sorted(subj_files)


def process(project, datadir):
    file_field = 'arc_testfile'
    results = []
    def_field = project.def_field
    fields = [def_field, file_field, 'arc_response_date', 'arc_day_index', 'arc_order_index', 'vitals_date', 'date_devices_given']
    subj2id = {}
    subjects = []
    file_glob = f'{datadir}/device_*_test_*.json'
    uploaded_count = 0

    # Handle secondary ID
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        subj2id = {x[sec_field]: x[def_field] for x in rec if x[sec_field]}
        subjects = sorted(list(set([x[sec_field] for x in rec if x[sec_field]])))
    else:
        rec = project.export_records(fields=[def_field])
        subj2id = {x[def_field]: x[def_field] for x in rec if x[def_field]}
        subjects = sorted(list(set([x[def_field] for x in rec])))

    # Get file records
    all_records = project.export_records(fields=fields)
    arc_records = [x for x in all_records if x['redcap_repeat_instance']]

    all_uploaded = list(set([x[file_field] for x in all_records if x[file_field]]))

    # Get the file list
    files = sorted(glob.glob(file_glob))

    file_count = len(files)
    if file_count <= 0:
        logger.debug(f'no files found:{file_glob}')
        return []
    else:
        logger.debug(f'found {file_count} files:{file_glob}')

    # Process each subject
    for subj in subjects:

        subj_files = _subject_files(subj, files)

        if len(subj_files) == 0:
            continue

        subj_id = subj2id[subj]
        subj_events = list(set([x['redcap_event_name'] for x in all_records if x[def_field] == subj_id]))
        subj_records = [x for x in arc_records if x[def_field] == subj_id]
        subj_uploaded = list(set([x[file_field] for x in subj_records if x[file_field]]))
        uploaded_count += len(subj_uploaded)

        for subj_file in subj_files:
            base_file = os.path.basename(subj_file)
            test_record = None
            same_event = None

            if base_file in all_uploaded:
                logger.debug(f'already uploaded:{subj}:{base_file}')
                continue

            # Load file data
            data = _load_testfile(subj_file)

            if data['id'] == '0':
                # Ignore the practice tests
                logger.debug(f'skipping practice test:{base_file}')
                continue

            # Get params to match from file
            arc_response_date = datetime.utcfromtimestamp(data['session_date']).strftime('%Y-%m-%d')
            arc_day_index = data['day']
            arc_order_index = data['id']

            # Find existing record
            for r in subj_records:
                if not r['arc_response_date'] or (r['arc_response_date'] != arc_response_date):
                    # wrong date
                    continue
                elif str(r['redcap_repeat_instance']) != str(arc_order_index):
                    # wrong order
                    continue
                else:
                    # matches
                    test_record = r
                    break

            if not test_record:
                # try to match with similar records date
                for r in all_records:

                    if r[def_field] != subj_id:
                        continue


                    if not (r.get('arc_response_date', False) or r.get('date_devices_given', False) or r.get('vitals_date', False)):
                        continue

                    if r.get('arc_response_date', False) and abs((datetime.strptime(r['arc_response_date'], '%Y-%m-%d') - datetime.strptime(arc_response_date, '%Y-%m-%d')).days) > 4:
                        # wrong date
                        continue

                    if r.get('date_devices_given', False):
                        date_devices_given = datetime.strptime(r['date_devices_given'], '%Y-%m-%d')
                        session_date = datetime.strptime(arc_response_date, '%Y-%m-%d')
                        diff_days = abs((date_devices_given - session_date).days)
                        if diff_days > 14:
                            continue

                    if r.get('vitals_date', False) and abs((datetime.strptime(r['vitals_date'], '%Y-%m-%d') - datetime.strptime(arc_response_date, '%Y-%m-%d')).days) > 14:
                        continue
                    else:
                        same_event = r['redcap_event_name']
                        break

                # now match event instead of date
                for r in subj_records:
                    if str(r['redcap_event_name']) != str(same_event):
                        # wrong event
                        continue
                    elif str(r['arc_day_index']) != str(arc_day_index):
                        # wrong day
                        continue
                    elif str(r['arc_order_index']) != str(arc_order_index):
                        # wrong order
                        continue
                    else:
                        # matches
                        test_record = r
                        break

            if not test_record:
                # no match found, make a new one, but first determine event
                logger.debug(f'no record yet:{subj}:{arc_response_date}:{arc_day_index}:{arc_order_index}:{base_file}')

                event_id = None

                if same_event:
                    event_id = same_event
                #else:
                #    if 'month_24_arm_3' in subj_events:
                #        event_id = 'month_24_arm_3'
                #    elif 'month_16_arm_3' in subj_events:
                #        event_id = 'month_16_arm_3'
                #    elif 'month_8_arm_3' in subj_events:
                #        event_id = 'month_8_arm_3'
                #    elif 'baselinemonth_0_arm_3' in subj_events:
                #        event_id = 'baselinemonth_0_arm_3'

                if not event_id:
                    # no event found, cannot upload
                    logger.debug(f'no event found:{subj}:{arc_response_date}:{arc_day_index}:{arc_order_index}:{base_file}')
                    continue

                test_record = {
                    def_field: subj_id,
                    'redcap_event_name': event_id,
                    'redcap_repeat_instance': arc_order_index,
                    'redcap_repeat_instrument': 'arc_data',
                    'arc_testfile': '',
                }

          

            record_id = test_record[def_field]
            event_id = test_record['redcap_event_name']
            repeat_id = test_record['redcap_repeat_instance']
            instrument = test_record['redcap_repeat_instrument']

            if event_id.startswith('unscheduledad_hoc_arm_3') or event_id.startswith('screening'):
                logger.debug(f'skipping:{record_id}:{event_id}:{repeat_id}')
                continue

            logger.debug(f'uploading:{record_id}:{event_id}:{instrument}:{repeat_id}:{base_file}')


            logger.debug(f'sanity check:{record_id}:{event_id}:{instrument}:{repeat_id}:{base_file}')
            found = False
            for x in all_records:
                if x[def_field] != subj_id or x['redcap_event_name'] != event_id or x['redcap_repeat_instrument'] != 'arc_data' or x['redcap_repeat_instance'] != repeat_id:
                    continue

                if x[file_field]:
                    logger.debug(f'found:{record_id}:{event_id}:{instrument}:{repeat_id}:{base_file}')
                    found = True

            if found:
                continue

            # Upload file to redcap
            try:
                result = upload_file(
                    project,
                    record_id,
                    file_field,
                    subj_file,
                    event_id=event_id,
                    repeat_id=repeat_id)

                logger.info(f'uploaded:{subj}:{event_id}:{repeat_id}:{base_file}')
            except (ValueError) as err:
                logger.error(f'error uploading:{base_file}:{err}')

            if not result:
                logger.error(f'upload failed:{subj}:{event_id}')
                continue

            results.append({
                'result': 'COMPLETE',
                'description': 'arc_file2redcap',
                'category': 'arc_file2redcap',
                'subject': subj,
                'event': event_id,
                'repeat': repeat_id,
                'field': file_field})

    logger.debug(f'total number uploaded={uploaded_count}')

    return results



import logging
import sys
import redcap
from garjus.automations.etl_nihtoolbox_drnair import process


# For testing, we create a connection and run it.
# In production, process_project will be run in garjus.update.automations


if __name__ == "__main__":
    logging.basicConfig(
        format='%(asctime)s - %(levelname)s:%(module)s:%(message)s',
        level=logging.DEBUG,
        datefmt='%Y-%m-%d %H:%M:%S')

    logging.info('connecting to redcap')
    api_url = 'https://redcap.vanderbilt.edu/api/'
    api_key = sys.argv[1]

    logging.info('Running it')
    results = process(redcap.Project(api_url, api_key))

    logging.info(results)
    logging.info('Done!')



ACTIVITY_RENAME = {
    'redcap_repeat_instance': 'ID',
    'activity_description': 'DESCRIPTION',
    'activity_datetime': 'DATETIME',
    'activity_event': 'EVENT',
    'activity_repeat': 'REPEAT',
    'activity_field': 'FIELD',
    'activity_result': 'RESULT',
    'activity_scan': 'SCAN',
    'activity_subject': 'SUBJECT',
    'activity_session': 'SESSION',
    'activity_type': 'CATEGORY',
}

ANALYSES_RENAME = {
    'redcap_repeat_instance': 'ID',
    'analysis_name': 'NAME',
    'analysis_lead': 'INVESTIGATOR',
    'analysis_include': 'SUBJECTS',
    'analysis_processor': 'PROCESSOR',
    'analysis_input': 'INPUT',
    'analysis_output': 'OUTPUT',
    'analyses_complete': 'COMPLETE',
    'analysis_status': 'STATUS',
    'analysis_covars': 'COVARS',
    'analysis_notes': 'NOTES',
}

ISSUES_RENAME = {
    'redcap_repeat_instance': 'ID',
    'issue_date': 'DATETIME',
    'issue_description': 'DESCRIPTION',
    'issue_event': 'EVENT',
    'issue_field': 'FIELD',
    'issue_scan': 'SCAN',
    'issue_session': 'SESSION',
    'issue_subject': 'SUBJECT',
    'issue_type': 'CATEGORY',
}

PROCESSING_RENAME = {
    'redcap_repeat_instance': 'ID',
    'processor_file': 'FILE',
    'processor_filter': 'FILTER',
    'processor_args': 'ARGS',
    'processing_complete': 'COMPLETE',
}

TASKS_RENAME = {
    'task_assessor': 'ASSESSOR',
    'task_status': 'STATUS',
    'task_inputlist': 'INPUTLIST',
    'task_var2val': 'VAR2VAL',
    'task_memreq': 'MEMREQ',
    'task_walltime': 'WALLTIME',
    'task_procdate': 'PROCDATE',
    'task_timeused': 'TIMEUSED',
    'task_memused': 'MEMUSED',
    'task_yamlfile': 'YAMLFILE',
    'task_userinputs': 'USERINPUTS',
    'task_failcount': 'FAILCOUNT',
    'task_yamlupload': 'YAMLUPLOAD',
}

REPORTS_RENAME = {
    'redcap_repeat_instance': 'ID',
    'progress_name': 'NAME',
    'progress_datetime': 'DATE',
    'progress_pdf': 'PDF',
    'progress_zip': 'DATA',
    'double_resultspdf': 'PDF',
    'double_resultsfile': 'DATA',
    'double_name': 'NAME',
    'double_datetime': 'DATE',
    'ndaimage03_name': 'NAME',
    'ndaimage03_csvfile': 'DATA',
}


COLUMNS = {
    'activity': [
        'PROJECT', 'SUBJECT', 'SESSION', 'SCAN', 'ID', 'DESCRIPTION',
        'DATETIME', 'EVENT', 'REPEAT', 'FIELD', 'CATEGORY', 'RESULT', 'STATUS', 'SOURCE'],
    'assessors': [
        'PROJECT', 'SUBJECT', 'SESSION', 'SESSTYPE', 'NOTE', 'DATE', 'SITE',
        'ASSR', 'PROCSTATUS', 'PROCTYPE', 'JOBDATE', 'TIMEUSED', 'MEMUSED',
        'QCSTATUS', 'QCDATE', 'QCBY', 'XSITYPE', 'INPUTS', 'MODALITY', 'full_path'],
    'issues': [
        'ID', 'DATETIME', 'PROJECT', 'CATEGORY',
        'SUBJECT', 'SESSION', 'SCAN ', 'DESCRIPTION',
        'EVENT', 'FIELD', 'STATUS'],
    'scans': [
        'PROJECT', 'SUBJECT', 'SESSION', 'SESSTYPE', 'TRACER', 'NOTE', 'DATE', 'SITE',
        'DURATION', 'FRAMES', 'TR', 'THICK', 'SENSE', 'MB',
        'SCANID', 'SCANTYPE', 'QUALITY', 'RESOURCES', 'MODALITY', 'XSITYPE', 'full_path'],
    'processing': [
        'ID', 'PROJECT', 'TYPE', 'FILTER', 'FILE', 'CUSTOM', 'ARGS', 'YAMLUPLOAD', 'EDIT', 'COMPLETE'],
    'subjects': [
        'PROJECT', 'SUBJECT', 'AGE', 'SEX', 'RACE'],
    'tasks': [
        'ID', 'IDLINK', 'PROJECT', 'STATUS', 'PROCTYPE', 'MEMREQ', 'WALLTIME',
        'TIMEUSED', 'MEMUSED', 'ASSESSOR', 'PROCDATE', 'INPUTLIST', 'VAR2VAL',
        'IMAGEDIR', 'JOBTEMPLATE', 'YAMLFILE', 'YAMLUPLOAD', 'USERINPUTS', 'FAILCOUNT', 'USER'],
    'analyses': ['PROJECT', 'ID', 'NAME', 'STATUS', 'EDIT', 'NOTES', 'SUBJECTS', 'PROCESSOR', 'INVESTIGATOR', 'OUTPUT'],
    'processors': ['ID', 'PROJECT', 'TYPE', 'EDIT', 'FILE', 'FILTER', 'ARGS'],
    'sgp': ['PROJECT', 'SUBJECT', 'ASSR', 'PROCSTATUS', 'PROCTYPE', 'QCSTATUS', 'INPUTS', 'DATE', 'XSITYPE'],
    'reports': ['PROJECT', 'TYPE', 'ID', 'VIEW', 'NAME', 'DATE', 'PDF', 'DATA'],
}


PROCLIB = {
    'AMYVIDQA_v2': {
        'short_descrip': 'Regional Amyloid SUVR using cerebellum as reference.',
        'inputs_descrip': 'T1w MRI processed with FreeSurfer (FS7_v1), Amyloid PET',
        'procurl': 'https://github.com/ccmvumc/AMYVIDQA',
        'stats_subset': ['compositegm_suvr', 'cblmtot_suvr', 'cblmwm_suvr', 'cblmgm_suvr', 'hippocampus_suvr']
    },
    'BFC_v2': {
        'short_descrip': 'Basal Forebrain Volumes.',
        'inputs_descrip': 'T1w MRI',
        'procurl': 'https://github.com/ccmvumc/BFC',
        'stats_subset': ['CH4_L_VOL', 'CH4_R_VOL']
    },
    'BrainAgeGap_v2': {
        'short_descrip': 'Predicted age of brain.',
        'inputs_descrip': 'T1w MRI parcellated with BrainColor atlas',
        'procurl': 'https://pubmed.ncbi.nlm.nih.gov/32948749/',
    },
    'DnSeg_v1': {
        'short_descrip': 'Basal Forebrain labeling.',
        'inputs_descrip': 'T1w MRI',
        'procurl': 'https://github.com/DerekDoss/DnSeg',
    },
    'FALLYPRIDEQA_v2':{
        'short_descrip': 'Fallypride QA with Regional SUVR using whole cerebellum as reference.',
        'inputs_descrip': 'T1w MRI processed with FreeSurfer (FS7_v1), Fallypride PET',
        'procurl': 'https://github.com/bud42/FALLYPRIDEQA',
        'stats_subset': ['antcing_suvr', 'compositegm_suvr', 'cblmgm_suvr', 'cblmwm_suvr', 'cblmtot_suvr'],
    },
    'fmri_bct_v2': {
        'short_descrip': 'Brain Connectivity Toolbox measures.',
        'inputs_descrip': 'Resting MRI processed with fmri_roi_v2',
        'procurl': 'https://github.com/REMBRANDT-study/fmri_bct',
        'stats_subset': ['Shen268_thr0p3_degree', 'Schaefer400_thr0p3_degree'],
    },
    'fmri_msit_v2': {
        'short_descrip': 'fMRI MSIT task pre-processing and 1st-Level analysis.',
        'inputs_descrip': 'T1w MRI, MSIT fMRI, E-prime EDAT',
        'procurl': 'https://github.com/REMBRANDT-study/fmri_msit',
        'stats_subset': ['con_amyg_mean', 'inc_amyg_mean', 'med_pct_outliers', 'con_bnst_mean', 'inc_bnst_mean'],
    },
    'fmri_msit_v4': {
        'short_descrip': 'fMRI MSIT task pre-processing and 1st-Level analysis.',
        'inputs_descrip': 'T1w MRI, MSIT fMRI, E-prime EDAT',
        'procurl': 'https://github.com/REMBRANDT-study/fmri_msit',
        'stats_subset': ['amyg', 'antins', 'ba46', 'bnst', 'dacc', 'pcc', 'postins', 'pvn', 'sgacc', 'vmpfc'],
    },
    'fmri_rest_v4': {
        'short_descrip': 'fMRI Resting State pre-processing.',
        'inputs_descrip': 'T1w MRI, Resting State fMRI',
        'procurl': 'https://github.com/REMBRANDT-study/fmri_rest',
    },
    'fmri_roi_v2': {
        'short_descrip': 'Regional measures of functional connectivity.',
        'inputs_descrip': 'Resting State fMRI processed with fmri_rest_v2',
        'procurl': 'https://github.com/REMBRANDT-study/fmri_roi',
    },
    'FS7sclimbic_v0': {
        'short_descrip': 'FreeSurfer 7 ScLimbic - volumes of subcortical limbic regions including Basal Forebrain.',
        'inputs_descrip': 'T1w MRI processed with FreeSurfer (FS7_v1)',
        'procurl': 'https://surfer.nmr.mgh.harvard.edu/fswiki/ScLimbic',
        'stats_subset': ['Left-Basal-Forebrain', 'Right-Basal-Forebrain'],
    },
    'FEOBVQA_v2': {
        'short_descrip': 'Regional SUVR using Supra-ventricular White Matter as reference.',
        'inputs_descrip': 'T1w MRI processed with FreeSurfer (FS7_v1), FEOBV PET',
        'procurl': 'https://github.com/ccmvumc/FEOBVQA',
        'stats_subset': ['cblmwm_suvr', 'compositegm_suvr', 'cblmgm_suvr'],
    },
    'NMQA_v2': {
        'short_descrip': 'Neuromelanin contrast ratio (CR), Substantia Nigra (SN) to Crus Cerebri (CC).',
        'inputs_descrip': 'T1w MRI, Neuromelanin MRI',
        'procurl': 'https://github.com/bud42/NMQA',
    },
    'NMQA_v3': {
        'short_descrip': 'Neuromelanin contrast ratio (CR), Substantia Nigra (SN) to Crus Cerebri (CC).',
        'inputs_descrip': 'T1w MRI, Neuromelanin MRI',
        'procurl': 'https://github.com/bud42/NMQA',
    },
    'assemblynet_v1': {
        'short_descrip': 'AssemblyNet whole brain parcellation.',
        'inputs_descrip': 'T1w MRI',
        'procurl': 'https://github.com/volBrain/AssemblyNet',
    },
    'FS7_v1': {
        'short_descrip': 'FreeSurfer 7 recon-all - whole brain parcellation of cortical and sub-cortical.\n\
\n\
Outputs for cortical regions are volumes, thickness average, and surface area of left and right hemisphere for 34 parcels:\n\
bankssts, caudal anterior cingulate, caudal middle frontal, cuneus, entorhinal, frontal pole, \
fusiform, inferior parietal, inferior temporal, insula, isthmus cingulate, lateral occipital, \
lateral orbitofrontal, lingual, medial orbitofrontal, middle temporal, paracentral, \
parahippocampal, pars opercularis, pars orbitalis, pars triangularis, pericalcarine, \
postcentral, posterior cingulate, precentral, precuneus, rostral anterior cingulate, \
rostral middle frontal, superiorfrontal, superior parietal, superior temporal, \
supramarginal, temporal pole, transverse temporal.\n\
\n\
Outputs for subcortical structures are volumes for left and right hemispheres of 8 regions:\n\
accumbens area, amygdala, caudate, hippocampus, pallidum, putamen, thalamus, and \
ventraldc.\n\
\n\
The white matter hyperinsity volume (wmh) is whole brain.\n\
\n\
See examples.\n',
        'inputs_descrip': 'T1w MRI\n',
        'procurl': 'https://github.com/bud42/FS7',
        'stats_subset': [
            'amygdala_lh_volume',
            'amygdala_rh_volume',
            'accumbensarea_lh_volume',
            'accumbensarea_rh_volume',
            'caudate_lh_volume',
            'caudate_rh_volume',
            'hippocampus_lh_volume',
            'hippocampus_rh_volume',
            'pallidum_lh_volume',
            'pallidum_rh_volume',
            'putamen_lh_volume',
            'putamen_rh_volume',
            'thalamus_lh_volume',
            'thalamus_rh_volume',
            'entorhinal_lh_surfarea',
            'entorhinal_rh_surfarea',
            'entorhinal_lh_thickavg',
            'entorhinal_rh_thickavg',
            'entorhinal_lh_volume',
            'entorhinal_rh_volume',
            'frontalpole_lh_surfarea',
            'frontalpole_lh_thickavg',
            'frontalpole_lh_volume',
            'frontalpole_rh_surfarea',
            'frontalpole_rh_thickavg',
            'frontalpole_rh_volume',
            'insula_lh_surfarea',
            'insula_rh_surfarea',
            'insula_lh_thickavg',
            'insula_rh_thickavg',
            'insula_lh_volume',
            'insula_rh_volume',
            'superiorfrontal_lh_surfarea',
            'superiorfrontal_rh_surfarea',
            'superiorfrontal_lh_thickavg',
            'superiorfrontal_rh_thickavg',
            'superiorfrontal_lh_volume',
            'superiorfrontal_rh_volume',
        ],
    },
    'FS7HPCAMG_v1': {
        'short_descrip': 'FreeSurfer 7 hippocampus & amygdala sub-region volumes.',
        'inputs_descrip': 'T1w processed with FreeSurfer (FS7_v1)',
        'procurl': 'https://github.com/bud42/FS7HPCAMG_v1',
        'stats_subset': ['hpchead_lh', 'hpchead_rh', 'hpcbody_lh', 'hpcbody_rh', 'hpctail_lh', 'hpctail_rh'],
    },
    'LST_v1': {
        'short_descrip': 'Lesion Segmentation Toolbox - white matter lesion volumes.',
        'inputs_descrip': 'T1w MRI, FLAIR MRI',
        'procurl': 'https://github.com/ccmvumc/LST1',
    },
    'Multi_Atlas_v3': {
        'short_descrip': 'Multi Atlas Labeling with BrainColor atlas.',
        'inputs_descrip': 'T1w MRI',
        'procurl': 'https://pubmed.ncbi.nlm.nih.gov/27726243',
    },
    'SAMSEG_v1': {
        'short_descrip': 'Runs SAMSEG from FreeSurfer 7.2.',
        'inputs_descrip': 'T1w MRI processed with FreeSurfer (FS7_v1), FLAIR MRI',
        'procurl': 'https://surfer.nmr.mgh.harvard.edu/fswiki/Samseg',
    },
    'fmriqa_v4': {
        'inputs_descrip': 'T1w MRI processed with SLANT, fMRI',
        'short_descrip': 'Functional MRI QA',
        'procurl': 'https://github.com/baxpr/fmriqa',
        'stats_subset': ['dvars_mean', 'fd_mean'],
    },
    'fmri_emostroop_v2': {
        'short_descrip': 'fMRI EmoStroop Pre-processing and 1st-Level',
        'inputs_descrip': 'T1w MRI, fMRI, EDAT',
        'procurl': 'https://github.com/ccmvumc/fmri_emostroop:v2',
        'stats_subset': ['lhSFG2_incgtcon', 'rhSFG2_incgtcon', 'overall_rt_mean'],
    },
    'struct_preproc_v1': {
        'short_descrip': 'SPM Structural Pre-processing - segmentation/normalization',
        'inputs_descrip': 'T1w MRI, FLAIR MRI',
    },
    'FS7hypothal_v1': {
        'short_descrip': 'FreeSurfer Hypothalamic Subunit Segmentation',
        'inputs_descrip': 'T1w processed with FreeSurfer (FS7_v1)',
        'procurl': 'https://surfer.nmr.mgh.harvard.edu/fswiki/HypothalamicSubunits',
        'stats_subset': ['whole left', 'whole right'],
    },
    'examiner': {
        'short_descrip': 'NIH Examiner',
    },
    'msit_vitals': {
        'short_descrip': 'MSIT Vitals',
        'stats_subset': ['msit_pulse', 'rest_pulse', 'msit_sbp', 'rest_sbp', 'chg_sbp', 'chg_pulse'],
    },
    'dtiQA_synb0_v7': {
        'short_descrip': 'PreQual DTI pre-processing and Quality Report',
        'stats_subset': [
            'Uncinate_fasciculus_R_med_fa',
            'Uncinate_fasciculus_L_med_fa',
            'Cingulum_cingulate_gyrus_R_med_fa',
            'Cingulum_cingulate_gyrus_L_med_fa',
            'Cingulum_hippocampus_R_med_fa',
            'Cingulum__hippocampus_L_med_fa',
            'eddy_avg_abs_displacement',
            'eddy_avg_rel_displacement',
        ],
        'inputs_descrip': 'T1w MRI, Diffusion MRI',
        'procurl': 'https://github.com/MASILab/PreQual',
    },
}

STATLIB = {
    'BFC_v2': {
        'CH4_L_VOL': 'Basal Forebrain Left Hemisphere CH4 Volume',
        'CH4_R_VOL': 'Basal Forebrain Right Hemisphere CH4 Volume'
    },
    'BrainAgeGap_v2': {
        'bag_age_gap': 'Brain Age Gap',
        'bag_age_pred': 'Predicted Age',
    },
    'DnSeg_v1': {
        'nbm_lh': 'Nucleus Basalis of Meynert Left Hemisphere Volume',
        'nbm_rh': 'Nucleus Basalis of Meynert Right Hemisphere Volume',
    },
    'FALLYPRIDEQA_v2': {
        'accumbens_suvr': 'Accumbens regional mean SUVR normalized by Whole Cerebellum',
        'amygdala_suvr': 'Amygdala regional mean SUVR normalized by Whole Cerebellum',
        'antcing_suvr': 'Anterior Cingulate regional mean SUVR normalized by Whole Cerebellum',
        'antflobe_suvr': 'Anterior Frontal Lobe mean SUVR normalized by Whole Cerebellum',
        'caudate_suvr': 'Caudate regional mean SUVR normalized by Whole Cerebellum',
        'cblmgm_suvr': 'Cerebellum Gray Matter regional mean SUVR normalized by Whole Cerebellum',
        'cblmtot_suvr': 'Cerebellum Total regional mean SUVR normalized by Whole Cerebellum',
        'cblmwm_suvr': 'Cerebellum White Matter regional mean SUVR normalized by Whole Cerebellum',
        'compositegm_suvr': 'Composite Gray Matter regional mean SUVR normalized by Whole Cerebellum',
        'cortwm_suvr': 'Cortical White Matter regional mean SUVR normalized by Whole Cerebellum',
        'hippocampus_suvr': 'Hippocampus regional mean SUVR normalized by Whole Cerebellum',
        'latplobe_suvr': 'Lateral Parietal Lobe regional mean SUVR normalized by Whole Cerebellum',
        'lattlobe_suvr': 'Lateral Temporal Lobe regional mean SUVR normalized by Whole Cerebellum',
        'mofc_suvr': 'Medial Orbito-frontal Cortex regional mean SUVR normalized by Whole Cerebellum',
        'pallidum_suvr': 'Pallidum regional mean SUVR normalized by Whole Cerebellum',
        'postcing_suvr': 'Posterior Cingulate regional mean SUVR normalized by Whole Cerebellum',
        'putamen_suvr': 'Putamen regional mean SUVR normalized by Whole Cerebellum',
        'thalamus_suvr': 'Thalamus regional mean SUVR normalized by Whole Cerebellum',
        'ventraldc_suvr': 'Ventral Diencephalon regional mean SUVR normalized by Whole Cerebellum',
    },
    'FS7_v1': {
        'hippocampus_lh_volume': 'Hippocampus Volume Left Hemisphere',
        'hippocampus_rh_volume': 'Hippocampus Volume Right Hemisphere',
        'amygdala_lh_volume': 'Amygdala Volume Left Hemisphere',
        'amygdala_rh_volume': 'Amygdala Volume Right Hemisphere',
        'superiorfrontal_lh_surfarea': 'Superior Frontal Surface Area Left Hemisphere',
        'superiorfrontal_rh_surfarea': 'Superior Frontal Surface Area Right Hemisphere',
        'superiorfrontal_lh_thickavg': 'Superior Frontal Thickness Average Left Hemisphere',
        'superiorfrontal_rh_thickavg': 'Superior Frontal Thickness Average Right Hemisphere',
        'superiorfrontal_lh_volume': 'Superior Frontal Volume Left Hemisphere',
        'superiorfrontal_rh_volume': 'Superior Frontal Volume Right Hemisphere',
    },
    'FEOBVQA_v2': {
        'antcing_suvr': 'Anterior Cingulate SUVR normalized by Supra-ventricular White Matter',
        'antflobe_suvr': 'Anterior Frontal Lobe SUVR normalized by Supra-ventricular White Matter',
        'cblmgm_suvr': 'Cerebellar Gray Matter SUVR normalized by Supra-ventricular White Matter',
        'cblmwm_suvr': 'Cerebellar White Matter SUVR normalized by Supra-ventricular White Matter',
        'compositegm_suvr': 'Composite Gray Matter SUVR normalized by Supra-ventricular White Matter',
        'cblmgm_suvr': 'Cerebellar Gray Matter SUVR normalized by Supra-ventricular White Matter',
        'cortwm_eroded_suvr': 'Eroded Cortical White Matter SUVR normalized by Supra-ventricular White Matter',
        'latplobe_suvr': 'Lateral Parietal Lobe SUVR normalized by Supra-ventricular White Matter',
        'lattlobe_suvr': 'Lateral Temporal Lobe SUVR normalized by Supra-ventricular White Matter',
        'postcing_suvr': 'Posterior Cingulate SUVR normalized by Supra-ventricular White Matter',
    },
    'SAMSEG_v1': {
        'samseg_lesions': 'whole brain White Matter Lesion Volume in cubic millimeters',
        'samseg_sbtiv': 'segmentation-based (estimated) Total Intracranial Volume in cubic millimeters',
    },
    'FS7HPCAMG_v1': {
        'hpcbody_lh': 'Hippocampus Body Left Hemisphere Volume in cubic millimeters',
        'hpcbody_rh': 'Hippocampus Body Right Hemisphere Volume in cubic millimeters',
        'hpchead_lh': 'Hippocampus Head Left Hemisphere Volume in cubic millimeters',
        'hpchead_rh': 'Hippocampus Head Right Hemisphere Volume in cubic millimeters',
        'hpctail_lh': 'Hippocampus Tail Left Hemisphere Volume in cubic millimeters',
        'hpctail_rh': 'Hippocampus Tail Right Hemisphere Volume in cubic millimeters',
    },
    'FS7sclimbic_v0': {
        'Left-Basal-Forebrain': 'Basal Forebrain Left Hemisphere Volume in cubic millimeters',
        'Right-Basal-Forebrain': 'Basal Forebrain Right Hemisphere Volume in cubic millimeters',
    },
    'LST_v1': {
        'wml_volume': 'White Matter Lesion Volume in milliliters',
    },
    'fmriqa_v4': {
        'dvars_mean': 'Derivative of Variance of voxels (measure of signal consistency)',
        'fd_mean': 'Framewise Displacement (measure of motion between adjacent volumes)',
    },
    'struct_preproc_v1': {
        'Volume1': 'Gray Matter',
        'Volume2': 'White Matter',
        'Volume3': 'CSF',
    },
    'fmri_bct_v2': {
        'Schaefer400_thr0p1_deg': 'Degree'
    },
    'Multi_Atlas_v3': {
        'ticv_mm3': 'Total intracranial volume in cubic millimeters',
    },
    'FS7hypothal_v1': {
        'whole left': 'whole left Hypothalamus',
        'whole right': 'whole right Hypothalamus',
    },
    'NMQA_v2': {
        'cr_mean': 'Contrast Ratio total mean',
        'cr_left': 'Contrast Ratio left hemisphere',
        'cr_right': 'Contrast Ratio right hemisphere',
    },
    'NMQA_v3': {
        'cr_mean': 'Contrast Ratio total mean',
        'cr_mean_lh': 'Contrast Ratio left hemisphere',
        'cr_mean_rh': 'Contrast Ratio right hemisphere',
        'cr_mean_sn1': 'Contrast Ratio Substantia Nigra region 1',
        'cr_mean_sn2': 'Contrast Ratio Substantia Nigra region 2',
        'cr_mean_sn3': 'Contrast Ratio Substantia Nigra region 3',
    },
    'assemblynet_v1': {
        'icv': 'Intracranial Cavity Volume',
    },
    'examiner': {

    },
    'fmri_msit_v4': {
        'amyg': 'mean contrast estimate in Amygdala',
        'antins': 'mean contrast estimate in Anterior Insula',
        'ba46': 'mean contrast estimate in Brodmann Area 46',
        'bnst': 'mean contrast estimate in Bed Nucleus of the Stria Terminalis  (BNST)',
        'dacc': 'mean contrast estimate in Dorsal Anterior Cingulate (dACC)',
        'pcc': 'mean contrast estimate in Posterior Cingulate Cortex (PCC)',
        'postins': 'mean contrast estimate in Posterior Insula',
        'pvn': 'mean contrast estimate in Para-Ventricular Nucleus (PVN)',
        'sgacc': 'mean contrast estimate in sub-genual Anterior Cingulate Cortex (sgACC)',
        'vmpfc': 'mean contrast estimate in ventromedial Prefrontal Cortex (vmPFC)',
    },
    'msit_vitals': {
        'msit_pulse': 'Pulse during stress task',
        'rest_pulse': 'Pulse before stress task',
        'chg_pulse': 'Change in Pulse during stress task (MSIT)',
        'msit_sbp': 'Systolic Blood Pressure during stress task',
        'rest_sbp': 'Systolic Blood Pressure before stress task',
        'chg_sbp': 'Change in Systolic Blood Pressure during stress task (MSIT)',
    },
}



import os
import tempfile
import json
import logging


def process_project(
    garjus,
    project,
    slicetiming,
    scantypes,
    sites=None
):
    # Get the scans, filter to only fmri_rest1,fmri_rest that start with 14*
    # download the JSON, upload to JSON_NoSliceTiming, add SliceTiming, upload
    # slicetiming is an list of floats
    results = []

    logging.debug(f'xnat_add_slicetiming:loading data:{project}')
    df = garjus.scans(projects=[project])

    # Filter
    df = df[df['SCANTYPE'].isin(scantypes)]
    if sites:
        df = df[df['SITE'].isin(sites)]

    # Check each scan
    for i, scan in df.iterrows():
        if 'JSON_MissingSliceTiming' in scan['RESOURCES']:
            continue

        full_path = scan['full_path']
        logging.debug(f'adding slicetiming:{full_path}')

        res = garjus.xnat().select(f'{full_path}/resources/JSON')

        files = res.files().get()
        if len(files) == 0:
            logging.debug(f'no JSON files found:{full_path}')
            continue
        elif len(files) > 1:
            logging.debug(f'too many JSON files found:{full_path}')
            continue

        src = files[0]

        with tempfile.TemporaryDirectory() as tmpdir:
            dst = os.path.join(tmpdir, src)
            res.file(src).get(dst)

            if has_slicetiming(dst):
                logging.debug(f'already has SliceTiming:{full_path}')
                continue

            # upload to no slicetiming
            new_res = res.parent().resource('JSON_MissingSliceTiming')
            if new_res.exists():
                logging.debug(f'JSON_MissingSliceTiming exists:{full_path}')
            else:
                logging.debug(f'saving to JSON_MissingSliceTiming:{full_path}')
                new_res.file(src).put(dst)

            # add it and upload
            add_slicetiming(dst, slicetiming)
            logging.debug(f'uploading to JSON:{full_path}:{dst}')
            res.file(src).put(dst, overwrite=True)

            results.append({
                'result': 'COMPLETE',
                'description': 'add slicetiming',
                'subject': scan['SUBJECT'],
                'session': scan['SESSION'],
                'scan': scan['SCANID']})

    return results


def add_slicetiming(jsonfile, slicetiming):
    with open(jsonfile, 'r+') as f:
        data = json.load(f)
        data['SliceTiming'] = slicetiming
        f.seek(0)
        json.dump(data, f, indent=4)


def has_slicetiming(jsonfile):
    with open(jsonfile, 'r') as f:
        return ('SliceTiming' in f.read())



import pandas as pd
from garjus import Garjus


def _load(project, data):
    # Load the data back to redcap
    try:
        _response = project.import_records(data)
        assert 'count' in _response
        return True
    except (AssertionError, Exception) as err:
        logger.error(err)
        return False



subj2hscrp = {}
subj2order = {}
id2subj = {}

# Get the redcap via garjus
g = Garjus()
rc = g.primary('D3')

# Load file data
df = pd.read_excel('/Users/boydb1/VUMC Plasma hsCRP Nov 2024.xlsx')
for i, row in df.iterrows():
    order = row['Assay order ']
    subject = row['Subject Number']
    hscrp = row['hsCRP (mg/L)']
    subj2hscrp[subject] =  str(hscrp)
    subj2order[subject] = str(order)

# Get the field names for main id and secondary id
def_field = rc.def_field
sec_field = rc.export_project_info()['secondary_unique_field']

# Load the subject numbers
if sec_field:
    rec = rc.export_records(fields=[def_field, sec_field])
    id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}
else:
    rec = rc.export_records(fields=[def_field])
    id2subj = {x[def_field]: x[def_field] for x in rec if x[def_field]}

# Get the plasma records
records = rc.export_records(fields=['plasma_hscrp', 'plasma_file', 'plasma_assayorder'])
records = [x for x in records if x['redcap_event_name'] in ['baseline_arm_1', 'baseline_arm_2']]
records = [x for x in records if not x['redcap_repeat_instrument']]

# Iterate looking for blanks
for r in records:
    record_id = r['record_id']

    if r.get('plasma_hscrp', False):
        print('existing', record_id)
        continue
 
    # Get the subject data from file data
    subj = id2subj[record_id]
    hscrp = subj2hscrp.get(subj, None)
    order = subj2order.get(subj, None)
    if not hscrp or not order:
        print('nope', subj)
        continue

    # Upload new data
    data = {
        rc.def_field: record_id,
        'redcap_event_name': r['redcap_event_name'],
        'plasma_hscrp': hscrp,
        'plasma_assayorder': order,
        'plasma_file': 'VUMC Plasma hsCRP Nov 2024.xlsx',
        'plasma_complete': '2',
    }
    _load(rc, [data])

print('done!')



import logging

from dash import dcc, html, dash_table as dt, Input, Output, callback
import dash_bootstrap_components as dbc

from .. import utils
from . import data
from ....dictionary import COLUMNS


logger = logging.getLogger('dashboard.reports')


def get_content():
    #columns = utils.make_columns(COLUMNS.get('reports'))
    columns = utils.make_columns([
        'PROJECT',
        'TYPE',
        'NAME',
        'VIEW',
        'DATE',
        'PDF',
        'DATA',
    ])

    # Format columns with links as markdown text
    for i, c in enumerate(columns):
        if c['name'] in ['VIEW']:
            columns[i]['type'] = 'text'
            columns[i]['presentation'] = 'markdown'

    content = [
        dbc.Row([
            dbc.Col(
                dbc.Button(
                    'Refresh Data',
                    id='button-reports-refresh',
                    outline=True,
                    color='primary',
                    size='sm',
                ),
            ),
            dbc.Col(
                dcc.Dropdown(
                    id='dropdown-reports-proj',
                    multi=True,
                    placeholder='Select Project(s)',
                ),
                width=3,
            ),
            dbc.Col(
                dcc.Dropdown(
                    id='dropdown-reports-type',
                    multi=True,
                    placeholder='Select Types(s)',
                ),
                width=3,
            ),
            dbc.Col(
                dcc.Dropdown(
                    value='Current',
                    id='dropdown-reports-time',
                    multi=False,
                    placeholder='Select Time',
                ),
                width=3,
            ),
        ]),
        dbc.Spinner(id="loading-reports-table", children=[
            dbc.Label('Loading...', id='label-reports-rowcount1'),
        ]),
        dt.DataTable(
            columns=columns,
            data=[],
            page_action='none',
            sort_action='native',
            id='datatable-reports',
            style_cell={
                'textAlign': 'center',
                'padding': '15px 5px 15px 5px',
                'height': 'auto',
            },
            style_header={
                'fontWeight': 'bold',
            },
            style_cell_conditional=[
                {'if': {'column_id': 'NAME'}, 'textAlign': 'left'},
            ],
            # Aligns the markdown cells, both vertical and horizontal
            css=[dict(selector="p", rule="margin: 0; text-align: center")],
        ),
        html.Label('0', id='label-reports-rowcount2')]

    return content


def load_reports(refresh=False):
    return data.load_data(refresh)


@callback(
    [
     Output('dropdown-reports-proj', 'options'),
     Output('dropdown-reports-type', 'options'),
     Output('dropdown-reports-time', 'options'),
     Output('datatable-reports', 'data'),
     Output('label-reports-rowcount1', 'children'),
     Output('label-reports-rowcount2', 'children'),
    ],
    [
     Input('dropdown-reports-proj', 'value'),
     Input('dropdown-reports-type', 'value'),
     Input('dropdown-reports-time', 'value'),
     Input('button-reports-refresh', 'n_clicks'),
    ])
def update_reports(
    selected_proj,
    selected_type,
    selected_time,
    n_clicks
):
    refresh = False

    logger.debug('update_all')

    # Load selected data with refresh if requested

    if utils.was_triggered('button-reports-refresh'):
        logger.debug(f'reports refresh:clicks={n_clicks}')
        refresh = True

    df = load_reports(refresh)

    # Get options
    projects, types, times = data.load_options(df)
    projects = utils.make_options(projects)
    types = utils.make_options(types)
    times = utils.make_options(times)

    logger.debug(f'loaded options:{projects}')
    logger.debug(f'loaded options:{types}')
    logger.debug(f'loaded options:{times}')

    # Apply filters
    df = data.filter_data(df, selected_proj, selected_type, selected_time)

    # Get the table data as one row per assessor
    records = df.reset_index().to_dict('records')

    # Format records
    for r in records:
        # Make view a link
        _link = r['VIEW']
        _text = 'view'
        r['VIEW'] = f'[{_text}]({_link})'

    # Count how many rows are in the table
    rowcount = '{} rows'.format(len(records))

    return [projects, types, times, records, rowcount, rowcount]
